# -*- coding: utf-8 -*-
import json
import logging
import math
from datetime import date, datetime, timezone

from nectar.constants import (
    STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF6,
    STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF20,
    STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF21,
)
from nectarbase import operations

from .account import Account
from .amount import Amount
from .blockchainobject import BlockchainObject
from .exceptions import ContentDoesNotExistsException, VotingInvalidOnArchivedPost
from .instance import shared_blockchain_instance
from .price import Price
from .utils import (
    construct_authorperm,
    formatTimeString,
    formatToTimeStamp,
    make_patch,
    resolve_authorperm,
)

log = logging.getLogger(__name__)


class Comment(BlockchainObject):
    """Read data about a Comment/Post in the chain

    :param str authorperm: identifier to post/comment in the form of
        ``@author/permlink``
    :param str tags: defines which api is used. Can be bridge, tags, condenser or database (default = bridge)
    :param Hive blockchain_instance: :class:`nectar.hive.Steem` instance to use when accessing a RPC


    .. code-block:: python

    >>> from nectar.comment import Comment
    >>> from nectar.account import Account
    >>> from nectar import Steem
    >>> stm = Steem()
    >>> acc = Account("gtg", blockchain_instance=stm)
    >>> authorperm = acc.get_blog(limit=1)[0]["authorperm"]
    >>> c = Comment(authorperm)
    >>> postdate = c["created"]
    >>> postdate_str = c.json()["created"]

    """

    type_id = 8

    def __init__(
        self,
        authorperm,
        api="bridge",
        observer="",
        full=True,
        lazy=False,
        blockchain_instance=None,
        **kwargs,
    ):
        self.full = full
        self.lazy = lazy
        self.api = api
        self.observer = observer
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if isinstance(authorperm, str) and authorperm != "":
            [author, permlink] = resolve_authorperm(authorperm)
            self["id"] = 0
            self["author"] = author
            self["permlink"] = permlink
            self["authorperm"] = authorperm
        elif isinstance(authorperm, dict) and "author" in authorperm and "permlink" in authorperm:
            authorperm["authorperm"] = construct_authorperm(
                authorperm["author"], authorperm["permlink"]
            )
            authorperm = self._parse_json_data(authorperm)
        super(Comment, self).__init__(
            authorperm,
            id_item="authorperm",
            lazy=lazy,
            full=full,
            blockchain_instance=blockchain_instance,
        )

    def _parse_json_data(self, comment):
        parse_times = [
            "active",
            "cashout_time",
            "created",
            "last_payout",
            "last_update",
            "updated",
            "max_cashout_time",
        ]
        for p in parse_times:
            if p in comment and isinstance(comment.get(p), str):
                comment[p] = formatTimeString(comment.get(p, "1970-01-01T00:00:00"))
        # Parse Amounts
        sbd_amounts = [
            "total_payout_value",
            "max_accepted_payout",
            "pending_payout_value",
            "curator_payout_value",
            "total_pending_payout_value",
            "promoted",
        ]
        for p in sbd_amounts:
            if p in comment and isinstance(comment.get(p), (str, list, dict)):
                value = comment.get(p, "0.000 %s" % (self.blockchain.backed_token_symbol))
                if (
                    isinstance(value, str)
                    and value.split(" ")[1] != self.blockchain.backed_token_symbol
                ):
                    value = value.split(" ")[0] + " " + self.blockchain.backed_token_symbol
                comment[p] = Amount(value, blockchain_instance=self.blockchain)

        if "community" not in comment:
            comment["community"] = ""

        # turn json_metadata into python dict
        meta_str = comment.get("json_metadata", "{}")
        if meta_str == "{}":
            comment["json_metadata"] = meta_str
        if isinstance(meta_str, (str, bytes, bytearray)):
            try:
                comment["json_metadata"] = json.loads(meta_str)
            except Exception:
                comment["json_metadata"] = {}

        comment["tags"] = []
        if isinstance(comment["json_metadata"], dict):
            if "tags" in comment["json_metadata"]:
                comment["tags"] = comment["json_metadata"]["tags"]
            if "community" in comment["json_metadata"]:
                comment["community"] = comment["json_metadata"]["community"]

        parse_int = [
            "author_reputation",
            "net_rshares",
        ]
        for p in parse_int:
            if p in comment and isinstance(comment.get(p), str):
                comment[p] = int(comment.get(p, "0"))

        if "active_votes" in comment:
            new_active_votes = []
            for vote in comment["active_votes"]:
                if "time" in vote and isinstance(vote.get("time"), str):
                    vote["time"] = formatTimeString(vote.get("time", "1970-01-01T00:00:00"))
                parse_int = [
                    "rshares",
                    "reputation",
                ]
                for p in parse_int:
                    if p in vote and isinstance(vote.get(p), str):
                        try:
                            vote[p] = int(vote.get(p, "0"))
                        except ValueError:
                            vote[p] = int(0)
                new_active_votes.append(vote)
            comment["active_votes"] = new_active_votes
        return comment

    def refresh(self):
        if self.identifier == "":
            return
        if not self.blockchain.is_connected():
            return
        [author, permlink] = resolve_authorperm(self.identifier)
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        if self.blockchain.rpc.get_use_appbase():
            from nectarapi.exceptions import InvalidParameters

            try:
                if self.api == "tags":
                    content = self.blockchain.rpc.get_discussion(
                        {"author": author, "permlink": permlink}, api="tags"
                    )
                elif self.api == "database":
                    content = self.blockchain.rpc.list_comments(
                        {"start": [author, permlink], "limit": 1, "order": "by_permlink"},
                        api="database",
                    )
                elif self.api == "bridge":
                    content = self.blockchain.rpc.get_post(
                        {"author": author, "permlink": permlink, "observer": self.observer},
                        api="bridge",
                    )
                elif self.api == "condenser":
                    content = self.blockchain.rpc.get_content(author, permlink, api="condenser")
                else:
                    raise ValueError("api must be: tags, database, bridge or condenser")
                if content is not None and "comments" in content:
                    content = content["comments"]
                if isinstance(content, list) and len(content) > 0:
                    content = content[0]
            except InvalidParameters:
                raise ContentDoesNotExistsException(self.identifier)
        else:
            from nectarapi.exceptions import InvalidParameters

            try:
                content = self.blockchain.rpc.get_content(author, permlink)
            except InvalidParameters:
                raise ContentDoesNotExistsException(self.identifier)
        if not content or not content["author"] or not content["permlink"]:
            raise ContentDoesNotExistsException(self.identifier)
        content = self._parse_json_data(content)
        content["authorperm"] = construct_authorperm(content["author"], content["permlink"])
        super(Comment, self).__init__(
            content,
            id_item="authorperm",
            lazy=self.lazy,
            full=self.full,
            blockchain_instance=self.blockchain,
        )

    def json(self):
        output = self.copy()
        if "authorperm" in output:
            output.pop("authorperm")
        if "json_metadata" in output:
            output["json_metadata"] = json.dumps(output["json_metadata"], separators=[",", ":"])
        if "tags" in output:
            output.pop("tags")
        parse_times = [
            "active",
            "cashout_time",
            "created",
            "last_payout",
            "last_update",
            "updated",
            "max_cashout_time",
        ]
        for p in parse_times:
            if p in output:
                p_date = output.get(p, datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    output[p] = formatTimeString(p_date)
                else:
                    output[p] = p_date
        sbd_amounts = [
            "total_payout_value",
            "max_accepted_payout",
            "pending_payout_value",
            "curator_payout_value",
            "total_pending_payout_value",
            "promoted",
        ]
        for p in sbd_amounts:
            if p in output and isinstance(output[p], Amount):
                output[p] = output[p].json()
        parse_int = [
            "author_reputation",
            "net_rshares",
        ]
        for p in parse_int:
            if p in output and isinstance(output[p], int):
                output[p] = str(output[p])
        if "active_votes" in output:
            new_active_votes = []
            for vote in output["active_votes"]:
                if "time" in vote:
                    p_date = vote.get("time", datetime(1970, 1, 1, 0, 0))
                    if isinstance(p_date, (datetime, date)):
                        vote["time"] = formatTimeString(p_date)
                    else:
                        vote["time"] = p_date
                parse_int = [
                    "rshares",
                    "reputation",
                ]
                for p in parse_int:
                    if p in vote and isinstance(vote[p], int):
                        vote[p] = str(vote[p])
                new_active_votes.append(vote)
            output["active_votes"] = new_active_votes
        return json.loads(str(json.dumps(output)))

    @property
    def id(self):
        return self["id"]

    @property
    def author(self):
        return self["author"]

    @property
    def permlink(self):
        return self["permlink"]

    @property
    def authorperm(self):
        return construct_authorperm(self["author"], self["permlink"])

    @property
    def category(self):
        if "category" in self:
            return self["category"]
        else:
            return ""

    @property
    def community(self):
        if "community" in self:
            return self["community"]
        else:
            return ""

    @property
    def community_title(self):
        """The Community title property."""
        if "community_title" in self:
            return self["community_title"]
        else:
            return ""

    @property
    def parent_author(self):
        if "parent_author" in self:
            return self["parent_author"]
        else:
            return ""

    @property
    def parent_permlink(self):
        if "parent_permlink" in self:
            return self["parent_permlink"]
        else:
            return ""

    @property
    def depth(self):
        return self["depth"]

    @property
    def title(self):
        if "title" in self:
            return self["title"]
        else:
            return ""

    @property
    def body(self):
        if "body" in self:
            return self["body"]
        else:
            return ""

    @property
    def json_metadata(self):
        if "json_metadata" in self:
            return self["json_metadata"]
        else:
            return {}

    def is_main_post(self):
        """Returns True if main post, and False if this is a comment (reply)."""
        if "depth" in self:
            return self["depth"] == 0
        else:
            return self["parent_author"] == ""

    def is_comment(self):
        """Returns True if post is a comment"""
        if "depth" in self:
            return self["depth"] > 0
        else:
            return self["parent_author"] != ""

    @property
    def reward(self):
        """Return the estimated total SBD reward."""
        a_zero = Amount(0, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain)
        author = Amount(self.get("total_payout_value", a_zero), blockchain_instance=self.blockchain)
        curator = Amount(
            self.get("curator_payout_value", a_zero), blockchain_instance=self.blockchain
        )
        pending = Amount(
            self.get("pending_payout_value", a_zero), blockchain_instance=self.blockchain
        )
        return author + curator + pending

    def is_pending(self):
        """Returns if the payout is pending (the post/comment
        is younger than 7 days)
        """
        a_zero = Amount(0, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain)
        total = Amount(self.get("total_payout_value", a_zero), blockchain_instance=self.blockchain)
        post_age_days = self.time_elapsed().total_seconds() / 60 / 60 / 24
        return post_age_days < 7.0 and float(total) == 0

    def time_elapsed(self):
        """Returns a timedelta on how old the post is."""
        return datetime.now(timezone.utc) - self["created"]

    def curation_penalty_compensation_SBD(self):
        """Returns The required post payout amount after 15 minutes
        which will compentsate the curation penalty, if voting earlier than 15 minutes
        """
        self.refresh()
        if self.blockchain.hardfork >= 21:
            reverse_auction_window_seconds = STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF21
        elif self.blockchain.hardfork >= 20:
            reverse_auction_window_seconds = STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF20
        else:
            reverse_auction_window_seconds = STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF6
        return (
            self.reward
            * reverse_auction_window_seconds
            / ((self.time_elapsed()).total_seconds() / 60) ** 2
        )

    def estimate_curation_SBD(self, vote_value_SBD, estimated_value_SBD=None):
        """Estimates curation reward

        :param float vote_value_SBD: The vote value in SBD for which the curation
            should be calculated
        :param float estimated_value_SBD: When set, this value is used for calculate
            the curation. When not set, the current post value is used.
        """
        self.refresh()
        if estimated_value_SBD is None:
            estimated_value_SBD = float(self.reward)
        t = 1.0 - self.get_curation_penalty()
        k = vote_value_SBD / (vote_value_SBD + float(self.reward))
        K = (1 - math.sqrt(1 - k)) / 4 / k
        return K * vote_value_SBD * t * math.sqrt(estimated_value_SBD)

    def get_curation_penalty(self, vote_time=None):
        """If post is less than 5 minutes old, it will incur a curation
        reward penalty.

        :param datetime vote_time: A vote time can be given and the curation
            penalty is calculated regarding the given time (default is None)
            When set to None, the current date is used.
        :returns: Float number between 0 and 1 (0.0 -> no penalty, 1.0 -> 100 % curation penalty)
        :rtype: float

        """
        if vote_time is None:
            elapsed_seconds = self.time_elapsed().total_seconds()
        elif isinstance(vote_time, str):
            elapsed_seconds = (formatTimeString(vote_time) - self["created"]).total_seconds()
        elif isinstance(vote_time, (datetime, date)):
            elapsed_seconds = (vote_time - self["created"]).total_seconds()
        else:
            raise ValueError("vote_time must be a string or a datetime")
        if self.blockchain.hardfork >= 21:
            reward = elapsed_seconds / STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF21
        elif self.blockchain.hardfork >= 20:
            reward = elapsed_seconds / STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF20
        else:
            reward = elapsed_seconds / STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF6
        if reward > 1:
            reward = 1.0
        return 1.0 - reward

    def get_vote_with_curation(self, voter=None, raw_data=False, pending_payout_value=None):
        """Returns vote for voter. Returns None, if the voter cannot be found in `active_votes`.

        :param str voter: Voter for which the vote should be returned
        :param bool raw_data: If True, the raw data are returned
        :param pending_payout_SBD: When not None this value instead of the current
            value is used for calculating the rewards
        :type pending_payout_SBD: float, str

        """
        specific_vote = None
        if voter is None:
            voter = Account(self["author"], blockchain_instance=self.blockchain)
        else:
            voter = Account(voter, blockchain_instance=self.blockchain)
        if "active_votes" in self:
            for vote in self["active_votes"]:
                if voter["name"] == vote["voter"]:
                    specific_vote = vote
        else:
            active_votes = self.get_votes()
            for vote in active_votes:
                if voter["name"] == vote["voter"]:
                    specific_vote = vote
        if specific_vote is not None and (raw_data or not self.is_pending()):
            return specific_vote
        elif specific_vote is not None:
            curation_reward = self.get_curation_rewards(
                pending_payout_SBD=True, pending_payout_value=pending_payout_value
            )
            specific_vote["curation_reward"] = curation_reward["active_votes"][voter["name"]]
            specific_vote["ROI"] = (
                float(curation_reward["active_votes"][voter["name"]])
                / float(voter.get_voting_value_SBD(voting_weight=specific_vote["percent"] / 100))
                * 100
            )
            return specific_vote
        else:
            return None

    def get_beneficiaries_pct(self):
        """Returns the sum of all post beneficiaries in percentage"""
        beneficiaries = self["beneficiaries"]
        weight = 0
        for b in beneficiaries:
            weight += b["weight"]
        return weight / 100.0

    def get_rewards(self):
        """Returns the total_payout, author_payout and the curator payout in SBD.
        When the payout is still pending, the estimated payout is given out.

        .. note:: Potential beneficiary rewards were already deducted from the
                  `author_payout` and the `total_payout`

        Example:::

            {
                'total_payout': 9.956 SBD,
                'author_payout': 7.166 SBD,
                'curator_payout': 2.790 SBD
            }

        """
        if self.is_pending():
            total_payout = Amount(self["pending_payout_value"], blockchain_instance=self.blockchain)
            author_payout = self.get_author_rewards()["total_payout_SBD"]
            curator_payout = total_payout - author_payout
        else:
            author_payout = Amount(self["total_payout_value"], blockchain_instance=self.blockchain)
            curator_payout = Amount(
                self["curator_payout_value"], blockchain_instance=self.blockchain
            )
            total_payout = author_payout + curator_payout
        return {
            "total_payout": total_payout,
            "author_payout": author_payout,
            "curator_payout": curator_payout,
        }

    def get_author_rewards(self):
        """Returns the author rewards.



        Example::

            {
                'pending_rewards': True,
                'payout_SP': 0.912 STEEM,
                'payout_SBD': 3.583 SBD,
                'total_payout_SBD': 7.166 SBD
            }

        """
        if not self.is_pending():
            return {
                "pending_rewards": False,
                "payout_SP": Amount(
                    0, self.blockchain.token_symbol, blockchain_instance=self.blockchain
                ),
                "payout_SBD": Amount(
                    0, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain
                ),
                "total_payout_SBD": Amount(
                    self["total_payout_value"], blockchain_instance=self.blockchain
                ),
            }
        author_reward_factor = 0.5
        median_hist = self.blockchain.get_current_median_history()
        if median_hist is not None:
            median_price = Price(median_hist, blockchain_instance=self.blockchain)
        beneficiaries_pct = self.get_beneficiaries_pct()
        curation_tokens = self.reward * author_reward_factor
        author_tokens = self.reward - curation_tokens
        curation_rewards = self.get_curation_rewards()
        if self.blockchain.hardfork >= 20 and median_hist is not None:
            author_tokens += median_price * curation_rewards["unclaimed_rewards"]

        benefactor_tokens = author_tokens * beneficiaries_pct / 100.0
        author_tokens -= benefactor_tokens

        if median_hist is not None and "percent_steem_dollars" in self:
            sbd_steem = author_tokens * self["percent_steem_dollars"] / 20000.0
            vesting_steem = median_price.as_base(self.blockchain.token_symbol) * (
                author_tokens - sbd_steem
            )
            return {
                "pending_rewards": True,
                "payout_SP": vesting_steem,
                "payout_SBD": sbd_steem,
                "total_payout_SBD": author_tokens,
            }
        elif median_hist is not None and "percent_hbd" in self:
            sbd_steem = author_tokens * self["percent_hbd"] / 20000.0
            vesting_steem = median_price.as_base(self.blockchain.token_symbol) * (
                author_tokens - sbd_steem
            )
            return {
                "pending_rewards": True,
                "payout_SP": vesting_steem,
                "payout_SBD": sbd_steem,
                "total_payout_SBD": author_tokens,
            }
        else:
            return {
                "pending_rewards": True,
                "total_payout": author_tokens,
                "payout_SBD": None,
                "total_payout_SBD": None,
            }

    def get_curation_rewards(self, pending_payout_SBD=False, pending_payout_value=None):
        """Returns the curation rewards. The split between creator/curator is currently 50%/50%.

        :param bool pending_payout_SBD: If True, the rewards are returned in SBD and not in STEEM (default is False)
        :param pending_payout_value: When not None this value instead of the current
            value is used for calculating the rewards
        :type pending_payout_value: float, str

        `pending_rewards` is True when
        the post is younger than 7 days. `unclaimed_rewards` is the
        amount of curation_rewards that goes to the author (self-vote or votes within
        the first 30 minutes). `active_votes` contains all voter with their curation reward.

        Example::

            {
                'pending_rewards': True, 'unclaimed_rewards': 0.245 STEEM,
                'active_votes': {
                    'leprechaun': 0.006 STEEM, 'timcliff': 0.186 STEEM,
                    'st3llar': 0.000 STEEM, 'crokkon': 0.015 STEEM, 'feedyourminnows': 0.003 STEEM,
                    'isnochys': 0.003 STEEM, 'loshcat': 0.001 STEEM, 'greenorange': 0.000 STEEM,
                    'qustodian': 0.123 STEEM, 'jpphotography': 0.002 STEEM, 'thinkingmind': 0.001 STEEM,
                    'oups': 0.006 STEEM, 'mattockfs': 0.001 STEEM, 'thecrazygm': 0.003 STEEM, 'michaelizer': 0.004 STEEM,
                    'flugschwein': 0.010 STEEM, 'ulisessabeque': 0.000 STEEM, 'hakancelik': 0.002 STEEM, 'sbi2': 0.008 STEEM,
                    'zcool': 0.000 STEEM, 'steemhq': 0.002 STEEM, 'rowdiya': 0.000 STEEM, 'qurator-tier-1-2': 0.012 STEEM
                }
            }

        """
        median_hist = self.blockchain.get_current_median_history()
        if median_hist is not None:
            median_price = Price(median_hist, blockchain_instance=self.blockchain)
        pending_rewards = False
        active_votes_list = self.get_votes()
        curator_reward_factor = 0.5

        if "total_vote_weight" in self:
            total_vote_weight = self["total_vote_weight"]
        active_votes_json_list = []
        for vote in active_votes_list:
            if "weight" not in vote:
                vote.refresh()
                active_votes_json_list.append(vote.json())
            else:
                active_votes_json_list.append(vote.json())

        total_vote_weight = 0
        for vote in active_votes_json_list:
            total_vote_weight += vote["weight"]

        if not self.is_pending():
            if pending_payout_SBD or median_hist is None:
                max_rewards = Amount(
                    self["curator_payout_value"], blockchain_instance=self.blockchain
                )
            else:
                max_rewards = median_price.as_base(self.blockchain.token_symbol) * Amount(
                    self["curator_payout_value"], blockchain_instance=self.blockchain
                )
            unclaimed_rewards = Amount(
                0, self.blockchain.token_symbol, blockchain_instance=self.blockchain
            )
        else:
            if pending_payout_value is None and "pending_payout_value" in self:
                pending_payout_value = Amount(
                    self["pending_payout_value"], blockchain_instance=self.blockchain
                )
            elif pending_payout_value is None:
                pending_payout_value = 0
            elif isinstance(pending_payout_value, (float, int)):
                pending_payout_value = Amount(
                    pending_payout_value,
                    self.blockchain.backed_token_symbol,
                    blockchain_instance=self.blockchain,
                )
            elif isinstance(pending_payout_value, str):
                pending_payout_value = Amount(
                    pending_payout_value, blockchain_instance=self.blockchain
                )
            if pending_payout_SBD or median_hist is None:
                max_rewards = pending_payout_value * curator_reward_factor
            else:
                max_rewards = median_price.as_base(self.blockchain.token_symbol) * (
                    pending_payout_value * curator_reward_factor
                )
            unclaimed_rewards = max_rewards.copy()
            pending_rewards = True

        active_votes = {}

        for vote in active_votes_json_list:
            if total_vote_weight > 0:
                claim = max_rewards * int(vote["weight"]) / total_vote_weight
            else:
                claim = 0
            if claim > 0 and pending_rewards:
                unclaimed_rewards -= claim
            if claim > 0:
                active_votes[vote["voter"]] = claim
            else:
                active_votes[vote["voter"]] = 0

        return {
            "pending_rewards": pending_rewards,
            "unclaimed_rewards": unclaimed_rewards,
            "active_votes": active_votes,
        }

    def get_reblogged_by(self, identifier=None):
        """Shows in which blogs this post appears"""
        if not identifier:
            post_author = self["author"]
            post_permlink = self["permlink"]
        else:
            [post_author, post_permlink] = resolve_authorperm(identifier)
        if not self.blockchain.is_connected():
            return None
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            return self.blockchain.rpc.get_reblogged_by(
                {"author": post_author, "permlink": post_permlink}, api="condenser"
            )["accounts"]
        else:
            return self.blockchain.rpc.get_reblogged_by(post_author, post_permlink, api="condenser")

    def get_replies(self, raw_data=False, identifier=None):
        """Returns content replies

        :param bool raw_data: When set to False, the replies will be returned as Comment class objects
        """
        if not identifier:
            post_author = self["author"]
            post_permlink = self["permlink"]
        else:
            [post_author, post_permlink] = resolve_authorperm(identifier)
        if not self.blockchain.is_connected():
            return None
        self.blockchain.rpc.set_next_node_on_empty_reply(False)

        # Use bridge.get_discussion API
        content_replies = self.blockchain.rpc.get_discussion(
            {"author": post_author, "permlink": post_permlink, "observer": self.observer},
            api="bridge",
        )

        if not content_replies:
            return []

        # The response format is a dict with keys in 'author/permlink' format
        # We need to extract the replies by filtering out the original post
        original_key = f"{post_author}/{post_permlink}"
        replies = []

        for key, content in content_replies.items():
            # Skip the original post
            if key == original_key:
                continue
            # Add the reply
            replies.append(content)

        if raw_data:
            return replies
        return [Comment(c, blockchain_instance=self.blockchain) for c in replies]

    def get_all_replies(self, parent=None):
        """Returns all content replies"""
        if parent is None:
            parent = self
        if parent["children"] > 0:
            children = parent.get_replies()
            if children is None:
                return []
            for cc in children[:]:
                children.extend(self.get_all_replies(parent=cc))
            return children
        return []

    def get_parent(self, children=None):
        """Returns the parent post with depth == 0"""
        if children is None:
            children = self
        while children["depth"] > 0:
            children = Comment(
                construct_authorperm(children["parent_author"], children["parent_permlink"]),
                blockchain_instance=self.blockchain,
            )
        return children

    def get_votes(self, raw_data=False):
        """Returns all votes as ActiveVotes object"""
        if raw_data and "active_votes" in self:
            return self["active_votes"]
        from .vote import ActiveVotes

        authorperm = construct_authorperm(self["author"], self["permlink"])
        return ActiveVotes(authorperm, lazy=False, blockchain_instance=self.blockchain)

    def upvote(self, weight=+100, voter=None):
        """Upvote the post

        :param float weight: (optional) Weight for posting (-100.0 -
            +100.0) defaults to +100.0
        :param str voter: (optional) Voting account

        """
        if weight < 0:
            raise ValueError("Weight must be >= 0.")
        last_payout = self.get("last_payout", None)
        if last_payout is not None:
            if formatToTimeStamp(last_payout) > 0:
                raise VotingInvalidOnArchivedPost
        return self.vote(weight, account=voter)

    def downvote(self, weight=100, voter=None):
        """Downvote the post

        :param float weight: (optional) Weight for posting (-100.0 -
            +100.0) defaults to -100.0
        :param str voter: (optional) Voting account

        """
        if weight < 0:
            raise ValueError("Weight must be >= 0.")
        last_payout = self.get("last_payout", None)
        if last_payout is not None:
            if formatToTimeStamp(last_payout) > 0:
                raise VotingInvalidOnArchivedPost
        return self.vote(-weight, account=voter)

    def vote(self, weight, account=None, identifier=None, **kwargs):
        """Vote for a post

        :param float weight: Voting weight. Range: -100.0 - +100.0.
        :param str account: (optional) Account to use for voting. If
            ``account`` is not defined, the ``default_account`` will be used
            or a ValueError will be raised
        :param str identifier: Identifier for the post to vote. Takes the
            form ``@author/permlink``.

        """
        if not identifier:
            identifier = construct_authorperm(self["author"], self["permlink"])

        return self.blockchain.vote(weight, identifier, account=account)

    def edit(self, body, meta=None, replace=False):
        """Edit an existing post

        :param str body: Body of the reply
        :param json meta: JSON meta object that can be attached to the
            post. (optional)
        :param bool replace: Instead of calculating a *diff*, replace
            the post entirely (defaults to ``False``)

        """
        if not meta:
            meta = {}
        original_post = self

        if replace:
            newbody = body
        else:
            newbody = make_patch(original_post["body"], body)
            if not newbody:
                log.info("No changes made! Skipping ...")
                return

        reply_identifier = construct_authorperm(
            original_post["parent_author"], original_post["parent_permlink"]
        )

        new_meta = {}
        if meta is not None:
            if bool(original_post["json_metadata"]):
                new_meta = original_post["json_metadata"]
                for key in meta:
                    new_meta[key] = meta[key]
            else:
                new_meta = meta

        return self.blockchain.post(
            original_post["title"],
            newbody,
            reply_identifier=reply_identifier,
            author=original_post["author"],
            permlink=original_post["permlink"],
            json_metadata=new_meta,
        )

    def reply(self, body, title="", author="", meta=None):
        """Reply to an existing post

        :param str body: Body of the reply
        :param str title: Title of the reply post
        :param str author: Author of reply (optional) if not provided
            ``default_user`` will be used, if present, else
            a ``ValueError`` will be raised.
        :param json meta: JSON meta object that can be attached to the
            post. (optional)

        """
        return self.blockchain.post(
            title, body, json_metadata=meta, author=author, reply_identifier=self.identifier
        )

    def delete(self, account=None, identifier=None):
        """Delete an existing post/comment

        :param str account: (optional) Account to use for deletion. If
            ``account`` is not defined, the ``default_account`` will be
            taken or a ValueError will be raised.

        :param str identifier: (optional) Identifier for the post to delete.
            Takes the form ``@author/permlink``. By default the current post
            will be used.

        .. note:: A post/comment can only be deleted as long as it has no
                  replies and no positive rshares on it.

        """
        if not account:
            if "default_account" in self.blockchain.config:
                account = self.blockchain.config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, blockchain_instance=self.blockchain)
        if not identifier:
            post_author = self["author"]
            post_permlink = self["permlink"]
        else:
            [post_author, post_permlink] = resolve_authorperm(identifier)
        op = operations.Delete_comment(**{"author": post_author, "permlink": post_permlink})
        return self.blockchain.finalizeOp(op, account, "posting")

    def resteem(self, identifier=None, account=None):
        """Resteem a post

        :param str identifier: post identifier (@<account>/<permlink>)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)

        """
        if not account:
            account = self.blockchain.configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, blockchain_instance=self.blockchain)
        if identifier is None:
            identifier = self.identifier
        author, permlink = resolve_authorperm(identifier)
        json_body = ["reblog", {"account": account["name"], "author": author, "permlink": permlink}]
        return self.blockchain.custom_json(
            id="follow", json_data=json_body, required_posting_auths=[account["name"]]
        )


class RecentReplies(list):
    """Obtain a list of recent replies

    :param str author: author
    :param bool skip_own: (optional) Skip replies of the author to him/herself.
        Default: True
    :param Steem blockchain_instance: Steem() instance to use when accesing a RPC
    """

    def __init__(
        self,
        author,
        skip_own=True,
        start_permlink="",
        limit=100,
        lazy=False,
        full=True,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if not self.blockchain.is_connected():
            return None
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        account = Account(author, blockchain_instance=self.blockchain)
        replies = account.get_account_posts(sort="replies", raw_data=True)
        comments = []
        if replies is None:
            replies = []
        for post in replies:
            if skip_own and post["author"] == author:
                continue
            comments.append(
                Comment(post, lazy=lazy, full=full, blockchain_instance=self.blockchain)
            )
        super(RecentReplies, self).__init__(comments)


class RecentByPath(list):
    """Obtain a list of posts recent by path, does the same as RankedPosts

    :param str path: path
    :param str tag: tag
    :param str observer: observer
    :param Steem blockchain_instance: Steem() instance to use when accesing a RPC
    """

    def __init__(
        self,
        path="trending",
        tag="",
        observer="",
        lazy=False,
        full=True,
        limit=20,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()

        # Create RankedPosts with proper parameters
        ranked_posts = RankedPosts(
            sort=path,
            tag=tag,
            observer=observer,
            limit=limit,
            lazy=lazy,
            full=full,
            blockchain_instance=self.blockchain,
            **kwargs,
        )

        super(RecentByPath, self).__init__(ranked_posts)


class RankedPosts(list):
    """Obtain a list of ranked posts

    :param str sort: can be: trending, hot, created, promoted, payout, payout_comments, muted
    :param str tag: tag, when used my, the community posts of the observer are shown
    :param str observer: Observer name
    :param int limit: limits the number of returns comments
    :param str start_author: start author
    :param str start_permlink: start permlink
    :param Steem blockchain_instance: Steem() instance to use when accesing a RPC
    """

    def __init__(
        self,
        sort,
        tag="",
        observer="",
        limit=21,
        start_author="",
        start_permlink="",
        lazy=False,
        full=True,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if not self.blockchain.is_connected():
            return None
        comments = []
        api_limit = limit
        if api_limit > 100:
            api_limit = 100
        last_n = -1
        while len(comments) < limit and last_n != len(comments):
            last_n = len(comments)
            self.blockchain.rpc.set_next_node_on_empty_reply(False)
            try:
                posts = self.blockchain.rpc.get_ranked_posts(
                    {
                        "sort": sort,
                        "tag": tag,
                        "observer": observer,
                        "limit": api_limit,
                        "start_author": start_author,
                        "start_permlink": start_permlink,
                    },
                    api="bridge",
                )
                if posts is None:
                    continue
                for post in posts:
                    if (
                        len(comments) > 0
                        and comments[-1]["author"] == post["author"]
                        and comments[-1]["permlink"] == post["permlink"]
                    ):
                        continue
                    if len(comments) >= limit:
                        continue
                    if raw_data:
                        comments.append(post)
                    else:
                        comments.append(
                            Comment(post, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                        )
                if len(comments) > 0:
                    start_author = comments[-1]["author"]
                    start_permlink = comments[-1]["permlink"]
                if limit - len(comments) < 100:
                    api_limit = limit - len(comments) + 1
            except Exception as e:
                # If we get an error but have some posts, return what we have
                if len(comments) > 0:
                    logging.warning(f"Error in RankedPosts: {str(e)}. Returning partial results.")
                    break
                # Otherwise, re-raise the exception
                raise
        super(RankedPosts, self).__init__(comments)


class AccountPosts(list):
    """Obtain a list of account related posts

    :param str sort: can be: comments, posts, blog, replies, feed
    :param str account: Account name
    :param str observer: Observer name
    :param int limit: limits the number of returns comments
    :param str start_author: start author
    :param str start_permlink: start permlink
    :param Hive blockchain_instance: Hive() instance to use when accesing a RPC
    """

    def __init__(
        self,
        sort,
        account,
        observer="",
        limit=20,
        start_author="",
        start_permlink="",
        lazy=False,
        full=True,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if not self.blockchain.is_connected():
            return None
        comments = []
        api_limit = limit
        if api_limit > 100:
            api_limit = 100
        last_n = -1
        while len(comments) < limit and last_n != len(comments):
            last_n = len(comments)
            self.blockchain.rpc.set_next_node_on_empty_reply(False)
            try:
                posts = self.blockchain.rpc.get_account_posts(
                    {
                        "sort": sort,
                        "account": account,
                        "observer": observer,
                        "limit": api_limit,
                        "start_author": start_author,
                        "start_permlink": start_permlink,
                    },
                    api="bridge",
                )
                if posts is None:
                    continue
                for post in posts:
                    if (
                        len(comments) > 0
                        and comments[-1]["author"] == post["author"]
                        and comments[-1]["permlink"] == post["permlink"]
                    ):
                        continue
                    if len(comments) >= limit:
                        continue
                    if raw_data:
                        comments.append(post)
                    else:
                        comments.append(
                            Comment(post, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                        )
                if len(comments) > 0:
                    start_author = comments[-1]["author"]
                    start_permlink = comments[-1]["permlink"]
                if limit - len(comments) < 100:
                    api_limit = limit - len(comments) + 1
            except Exception as e:
                # If we get an error but have some posts, return what we have
                if len(comments) > 0:
                    logging.warning(f"Error in AccountPosts: {str(e)}. Returning partial results.")
                    break
                # Otherwise, re-raise the exception
                raise
        super(AccountPosts, self).__init__(comments)



# -*- coding: utf-8 -*-
import unittest

from nectar import Steem
from nectar.account import Account
from nectar.hivesigner import HiveSigner

from .nodes import get_hive_nodes

# Py3 compatibility
core_unit = "STM"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(
            node=get_hive_nodes(),
            nobroadcast=True,
            unsigned=True,
            data_refresh_time_seconds=900,
            num_retries=10,
        )

        cls.account = Account("test", full=True, steem_instance=cls.bts)

    def test_transfer(self):
        bts = self.bts
        acc = self.account
        acc.blockchain.txbuffer.clear()
        tx = acc.transfer("test1", 1.000, "HIVE", memo="test")
        sc2 = HiveSigner(steem_instance=bts)
        url = sc2.url_from_tx(tx)
        url_test = (
            "https://hivesigner.com/sign/transfer?from=test&to=test1&amount=1.000+HIVE&memo=test"
        )
        self.assertEqual(len(url), len(url_test))
        self.assertEqual(len(url.split("?")), 2)
        self.assertEqual(url.split("?")[0], url_test.split("?")[0])

        url_parts = (url.split("?")[1]).split("&")
        url_test_parts = (url_test.split("?")[1]).split("&")

        self.assertEqual(len(url_parts), 4)
        self.assertEqual(len(list(set(url_parts).intersection(set(url_test_parts)))), 4)

    def test_login_url(self):
        bts = self.bts
        sc2 = HiveSigner(steem_instance=bts)
        url = sc2.get_login_url("localhost", scope="login,vote")
        url_test = "https://hivesigner.com/oauth2/authorize?client_id=None&redirect_uri=localhost&scope=login,vote"
        self.assertEqual(len(url), len(url_test))
        self.assertEqual(len(url.split("?")), 2)
        self.assertEqual(url.split("?")[0], url_test.split("?")[0])

        url_parts = (url.split("?")[1]).split("&")
        url_test_parts = (url_test.split("?")[1]).split("&")

        self.assertEqual(len(url_parts), 3)
        self.assertEqual(len(list(set(url_parts).intersection(set(url_test_parts)))), 3)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from nectar.account import Account
from nectar.steem import Steem

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    # stm = Steem(node="https://testnet.timcliff.com/")
    # stm = Steem(node="https://testnet.steemitdev.com")
    stm = Steem(node="https://api.steemit.com")
    stm.wallet.unlock(pwd="pwd123")

    account = Account("thecrazygm", steem_instance=stm)
    print(account.get_voting_power())

    account.transfer("thecrazygm", 0.001, "SBD", "test")



# -*- coding: utf-8 -*-
import logging

from .comment import Comment
from .instance import shared_blockchain_instance

log = logging.getLogger(__name__)


class Query(dict):
    """Query to be used for all discussion queries

    :param int limit: limits the number of posts
    :param str tag: tag query
    :param int truncate_body:
    :param array filter_tags:
    :param array select_authors:
    :param array select_tags:
    :param str start_author:
    :param str start_permlink:
    :param str start_tag:
    :param str parent_author:
    :param str parent_permlink:
    :param str start_parent_author:
    :param str before_date:
    :param str author: Author (see Discussions_by_author_before_date)

    .. testcode::

        from nectar.discussions import Query
        query = Query(limit=10, tag="steemit")

    """

    def __init__(
        self,
        limit=0,
        tag="",
        truncate_body=0,
        filter_tags=[],
        select_authors=[],
        select_tags=[],
        start_author=None,
        start_permlink=None,
        start_tag=None,
        parent_author=None,
        parent_permlink=None,
        start_parent_author=None,
        before_date=None,
        author=None,
    ):
        self["limit"] = limit
        self["truncate_body"] = truncate_body
        self["tag"] = tag
        self["filter_tags"] = filter_tags
        self["select_authors"] = select_authors
        self["select_tags"] = select_tags
        self["start_author"] = start_author
        self["start_permlink"] = start_permlink
        self["start_tag"] = start_tag
        self["parent_author"] = parent_author
        self["parent_permlink"] = parent_permlink
        self["start_parent_author"] = start_parent_author
        self["before_date"] = before_date
        self["author"] = author


class Discussions(object):
    """Get Discussions

    :param Steem blockchain_instance: Steem instance

    """

    def __init__(self, lazy=False, use_appbase=False, blockchain_instance=None, **kwargs):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.lazy = lazy
        self.use_appbase = use_appbase

    def get_discussions(self, discussion_type, discussion_query, limit=1000, raw_data=False):
        """Get Discussions

        :param str discussion_type: Defines the used discussion query
        :param Query discussion_query: Defines the parameter for
               searching posts
        :param bool raw_data: returns list of comments when False, default is False

        .. testcode::

            from nectar.discussions import Query, Discussions
            query = Query(limit=51, tag="steemit")
            discussions = Discussions()
            count = 0
            for d in discussions.get_discussions("tags", query, limit=200):
                print(("%d. " % (count + 1)) + str(d))
                count += 1

        """
        if limit >= 100 and discussion_query["limit"] == 0:
            discussion_query["limit"] = 100
        elif limit < 100 and discussion_query["limit"] == 0:
            discussion_query["limit"] = limit
        query_count = 0
        found_more_than_start_entry = True
        if "start_author" in discussion_query:
            start_author = discussion_query["start_author"]
        else:
            start_author = None
        if "start_permlink" in discussion_query:
            start_permlink = discussion_query["start_permlink"]
        else:
            start_permlink = None
        if "start_tag" in discussion_query:
            start_tag = discussion_query["start_tag"]
        else:
            start_tag = None
        if "start_parent_author" in discussion_query:
            start_parent_author = discussion_query["start_parent_author"]
        else:
            start_parent_author = None
        if not discussion_query["before_date"]:
            discussion_query["before_date"] = "1970-01-01T00:00:00"
        while query_count < limit and found_more_than_start_entry:
            rpc_query_count = 0
            dd = None
            discussion_query["start_author"] = start_author
            discussion_query["start_permlink"] = start_permlink
            discussion_query["start_tag"] = start_tag
            discussion_query["start_parent_author"] = start_parent_author
            if discussion_type == "trending":
                dd = Discussions_by_trending(
                    discussion_query, blockchain_instance=self.blockchain, lazy=self.lazy
                )
            elif discussion_type == "author_before_date":
                dd = Discussions_by_author_before_date(
                    author=discussion_query["author"],
                    start_permlink=discussion_query["start_permlink"],
                    before_date=discussion_query["before_date"],
                    limit=discussion_query["limit"],
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                )
            elif discussion_type == "payout":
                dd = Comment_discussions_by_payout(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "post_payout":
                dd = Post_discussions_by_payout(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "created":
                dd = Discussions_by_created(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "active":
                dd = Discussions_by_active(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "cashout":
                dd = Discussions_by_cashout(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "votes":
                dd = Discussions_by_votes(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "children":
                dd = Discussions_by_children(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "hot":
                dd = Discussions_by_hot(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "feed":
                dd = Discussions_by_feed(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "blog":
                dd = Discussions_by_blog(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "comments":
                dd = Discussions_by_comments(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "promoted":
                dd = Discussions_by_promoted(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "replies":
                dd = Discussions_by_replies(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                    raw_data=raw_data,
                )
            elif discussion_type == "tags":
                dd = Trending_tags(
                    discussion_query,
                    blockchain_instance=self.blockchain,
                    lazy=self.lazy,
                    use_appbase=self.use_appbase,
                )
            else:
                raise ValueError("Wrong discussion_type")
            if not dd:
                return

            for d in dd:
                double_result = False
                if discussion_type == "tags":
                    if query_count != 0 and rpc_query_count == 0 and (d["name"] == start_tag):
                        double_result = True
                        if len(dd) == 1:
                            found_more_than_start_entry = False
                    start_tag = d["name"]
                elif discussion_type == "replies":
                    if (
                        query_count != 0
                        and rpc_query_count == 0
                        and (d["author"] == start_parent_author and d["permlink"] == start_permlink)
                    ):
                        double_result = True
                        if len(dd) == 1:
                            found_more_than_start_entry = False
                    start_parent_author = d["author"]
                    start_permlink = d["permlink"]
                else:
                    if (
                        query_count != 0
                        and rpc_query_count == 0
                        and (d["author"] == start_author and d["permlink"] == start_permlink)
                    ):
                        double_result = True
                        if len(dd) == 1:
                            found_more_than_start_entry = False
                    start_author = d["author"]
                    start_permlink = d["permlink"]
                rpc_query_count += 1
                if not double_result:
                    query_count += 1
                    if query_count <= limit:
                        yield d


class Discussions_by_trending(list):
    """Get Discussions by trending

    :param Query discussion_query: Defines the parameter for
        searching posts
    :param Steem blockchain_instance: Steem instance
    :param bool raw_data: returns list of comments when False, default is False

    .. testcode::

        from nectar.discussions import Query, Discussions_by_trending
        q = Query(limit=10, tag="steem")
        for h in Discussions_by_trending(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "trending",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_trending(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_trending(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_trending, self).__init__([x for x in posts])
        else:
            super(Discussions_by_trending, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_author_before_date(list):
    """Get Discussions by author before date

    .. note:: To retrieve discussions before date, the time of creation
              of the discussion @author/start_permlink must be older than
              the specified before_date parameter.

    :param str author: Defines the author *(required)*
    :param str start_permlink: Defines the permlink of a starting discussion
    :param str before_date: Defines the before date for query
    :param int limit: Defines the limit of discussions
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_author_before_date
        for h in Discussions_by_author_before_date(limit=10, author="gtg"):
            print(h)

    """

    def __init__(
        self,
        author="",
        start_permlink="",
        before_date="1970-01-01T00:00:00",
        limit=100,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            if author:
                bridge_query = {
                    "sort": "posts",
                    "account": author,
                    "limit": limit,
                }
                if start_permlink:
                    bridge_query["start_permlink"] = start_permlink
                posts = self.blockchain.rpc.get_account_posts(bridge_query, api="bridge")
                # Filter by before_date if provided
                if before_date and before_date != "1970-01-01T00:00:00":
                    filtered_posts = []
                    for post in posts:
                        if "created" in post and post["created"] < before_date:
                            filtered_posts.append(post)
                    posts = filtered_posts
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                discussion_query = {
                    "author": author,
                    "start_permlink": start_permlink,
                    "before_date": before_date,
                    "limit": limit,
                }
                posts = self.blockchain.rpc.get_discussions_by_author_before_date(
                    discussion_query, api="tags"
                )["discussions"]
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_author_before_date(
                    author, start_permlink, before_date, limit
                )
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_author_before_date, self).__init__([x for x in posts])
        else:
            super(Discussions_by_author_before_date, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Comment_discussions_by_payout(list):
    """Get comment_discussions_by_payout

    :param Query discussion_query: Defines the parameter for
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Comment_discussions_by_payout
        q = Query(limit=10)
        for h in Comment_discussions_by_payout(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "payout_comments",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_comment_discussions_by_payout(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_comment_discussions_by_payout(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Comment_discussions_by_payout, self).__init__([x for x in posts])
        else:
            super(Comment_discussions_by_payout, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Post_discussions_by_payout(list):
    """Get post_discussions_by_payout

    :param Query discussion_query: Defines the parameter for
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Post_discussions_by_payout
        q = Query(limit=10)
        for h in Post_discussions_by_payout(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "payout",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_post_discussions_by_payout(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_post_discussions_by_payout(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Post_discussions_by_payout, self).__init__([x for x in posts])
        else:
            super(Post_discussions_by_payout, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_created(list):
    """Get discussions_by_created

    :param Query discussion_query: Defines the parameter for
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_created
        q = Query(limit=10)
        for h in Discussions_by_created(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "created",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_created(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_created(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_created, self).__init__([x for x in posts])
        else:
            super(Discussions_by_created, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_active(list):
    """get_discussions_by_active

    :param Query discussion_query: Defines the parameter
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem() instance to use when accesing a RPC

    .. testcode::

        from nectar.discussions import Query, Discussions_by_active
        q = Query(limit=10)
        for h in Discussions_by_active(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "active",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_active(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_active(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_active, self).__init__([x for x in posts])
        else:
            super(Discussions_by_active, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_cashout(list):
    """Get discussions_by_cashout. This query seems to be broken at the moment.
    The output is always empty.

    :param Query discussion_query: Defines the parameter
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_cashout
        q = Query(limit=10)
        for h in Discussions_by_cashout(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            # Note: 'payout' is the closest sort to 'cashout' in bridge API
            bridge_query = {
                "sort": "payout",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_cashout(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_cashout(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_cashout, self).__init__([x for x in posts])
        else:
            super(Discussions_by_cashout, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_votes(list):
    """Get discussions_by_votes

    :param Query discussion_query: Defines the parameter
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_votes
        q = Query(limit=10)
        for h in Discussions_by_votes(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            # Note: There is no direct 'votes' sort in bridge API, so we'll approximate using trending
            bridge_query = {
                "sort": "trending",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_votes(reduced_query, api="tags")[
                        "discussions"
                    ]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_votes(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_votes, self).__init__([x for x in posts])
        else:
            super(Discussions_by_votes, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_children(list):
    """Get discussions by children

    :param Query discussion_query: Defines the parameter
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_children
        q = Query(limit=10)
        for h in Discussions_by_children(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            # Note: There is no direct 'children' sort in bridge API, we'll use 'trending' as a fallback
            bridge_query = {
                "sort": "trending",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
            # We could try to sort posts by their children count here if needed
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_children(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_children(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_children, self).__init__([x for x in posts])
        else:
            super(Discussions_by_children, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_hot(list):
    """Get discussions by hot

    :param Query discussion_query: Defines the parameter
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_hot
        q = Query(limit=10, tag="steem")
        for h in Discussions_by_hot(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "hot",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_hot(reduced_query, api="tags")[
                        "discussions"
                    ]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_hot(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_hot, self).__init__([x for x in posts])
        else:
            super(Discussions_by_hot, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_feed(list):
    """Get discussions by feed

    :param Query discussion_query: Defines the parameter
        searching posts, tag musst be set to a username
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_feed
        q = Query(limit=10, tag="steemit")
        for h in Discussions_by_feed(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            account = reduced_query.get("tag", "")
            if account:
                bridge_query = {
                    "sort": "feed",
                    "account": account,
                    "limit": reduced_query.get("limit", 20),
                }
                if "start_author" in reduced_query and "start_permlink" in reduced_query:
                    bridge_query["start_author"] = reduced_query["start_author"]
                    bridge_query["start_permlink"] = reduced_query["start_permlink"]
                posts = self.blockchain.rpc.get_account_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_feed(reduced_query, api="tags")[
                        "discussions"
                    ]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_feed(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_feed, self).__init__([x for x in posts])
        else:
            super(Discussions_by_feed, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_blog(list):
    """Get discussions by blog

    :param Query discussion_query: Defines the parameter
        searching posts, tag musst be set to a username
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_blog
        q = Query(limit=10)
        for h in Discussions_by_blog(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            account = reduced_query.get("tag", "")
            if account:
                bridge_query = {
                    "sort": "blog",
                    "account": account,
                    "limit": reduced_query.get("limit", 20),
                }
                if "start_author" in reduced_query and "start_permlink" in reduced_query:
                    bridge_query["start_author"] = reduced_query["start_author"]
                    bridge_query["start_permlink"] = reduced_query["start_permlink"]
                posts = self.blockchain.rpc.get_account_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_blog(reduced_query, api="tags")
                    if isinstance(posts, dict) and "discussions" in posts:
                        posts = posts["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                self.blockchain.rpc.set_next_node_on_empty_reply(False)
                posts = self.blockchain.rpc.get_discussions_by_blog(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_blog, self).__init__([x for x in posts])
        else:
            super(Discussions_by_blog, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_comments(list):
    """Get discussions by comments

    :param Query discussion_query: Defines the parameter
        searching posts, start_author and start_permlink must be set.
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_comments
        q = Query(limit=10, start_author="steemit", start_permlink="firstpost")
        for h in Discussions_by_comments(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in ["start_author", "start_permlink", "limit"]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                # The bridge.get_discussion API retrieves an entire discussion tree
                author = reduced_query["start_author"]
                permlink = reduced_query["start_permlink"]
                bridge_query = {
                    "author": author,
                    "permlink": permlink,
                }
                # The bridge API returns a discussion tree, we need to flatten it
                discussion = self.blockchain.rpc.get_discussion(bridge_query, api="bridge")
                # Extract comments from the discussion tree
                if discussion and isinstance(discussion, dict):
                    posts = []
                    # Start with the main post
                    main_post = discussion.get(f"@{author}/{permlink}")
                    if main_post:
                        posts.append(main_post)
                    # Add replies
                    for key, value in discussion.items():
                        if key != f"@{author}/{permlink}" and isinstance(value, dict):
                            posts.append(value)
                    # Limit the number of posts if needed
                    if "limit" in reduced_query and len(posts) > reduced_query["limit"]:
                        posts = posts[: reduced_query["limit"]]
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_comments(
                        reduced_query, api="tags"
                    )
                    if "discussions" in posts:
                        posts = posts["discussions"]  # inconsistent format across node types
                except Exception:
                    posts = self.blockchain.rpc.get_discussions_by_comments(
                        reduced_query["start_author"],
                        reduced_query["start_permlink"],
                        reduced_query["limit"],
                    )
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_comments(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_comments, self).__init__([x for x in posts])
        else:
            super(Discussions_by_comments, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_promoted(list):
    """Get discussions by promoted

    :param Query discussion_query: Defines the parameter
        searching posts
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_promoted
        q = Query(limit=10, tag="steem")
        for h in Discussions_by_promoted(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in [
            "tag",
            "limit",
            "filter_tags",
            "select_authors",
            "select_tags",
            "truncate_body",
            "start_author",
            "start_permlink",
        ]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            bridge_query = {
                "sort": "promoted",
                "tag": reduced_query.get("tag", ""),
                "observer": "",
            }
            if "limit" in reduced_query:
                bridge_query["limit"] = reduced_query["limit"]
            if "start_author" in reduced_query and "start_permlink" in reduced_query:
                bridge_query["start_author"] = reduced_query["start_author"]
                bridge_query["start_permlink"] = reduced_query["start_permlink"]
            posts = self.blockchain.rpc.get_ranked_posts(bridge_query, api="bridge")
        except Exception:
            # Fall back to old API methods
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_discussions_by_promoted(
                        reduced_query, api="tags"
                    )["discussions"]
                except Exception:
                    posts = []
            if len(posts) == 0:
                posts = self.blockchain.rpc.get_discussions_by_promoted(reduced_query)
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_promoted, self).__init__([x for x in posts])
        else:
            super(Discussions_by_promoted, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Discussions_by_replies(list):
    """Get replies for an author's post

    :param Query discussion_query: Defines the parameter
        searching posts, start_parent_author, start_permlink must be set.
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Discussions_by_replies
        q = Query(limit=10, start_parent_author="steemit", start_permlink="firstpost")
        for h in Discussions_by_replies(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        reduced_query = {}
        for key in ["start_parent_author", "start_permlink", "limit"]:
            if key in discussion_query:
                reduced_query[key] = discussion_query[key]
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        try:
            # Try to use the bridge API first (preferred method)
            if "start_parent_author" in reduced_query and "start_permlink" in reduced_query:
                # The bridge.get_discussion API retrieves replies to a post as well
                author = reduced_query["start_parent_author"]
                permlink = reduced_query["start_permlink"]
                bridge_query = {
                    "author": author,
                    "permlink": permlink,
                }
                # The bridge API returns a discussion tree
                discussion = self.blockchain.rpc.get_discussion(bridge_query, api="bridge")
                # Extract replies from the discussion tree
                if discussion and isinstance(discussion, dict):
                    posts = []
                    # Gather all replies (all items except the main post)
                    main_post_key = f"@{author}/{permlink}"
                    for key, value in discussion.items():
                        if key != main_post_key and isinstance(value, dict):
                            posts.append(value)
                    # Limit the number of posts if needed
                    if "limit" in reduced_query and len(posts) > reduced_query["limit"]:
                        posts = posts[: reduced_query["limit"]]
        except Exception:
            # Fall back to old API methods
            posts = []
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                try:
                    posts = self.blockchain.rpc.get_replies_by_last_update(
                        reduced_query, api="tags"
                    )
                    if "discussions" in posts:
                        posts = posts["discussions"]
                except Exception:
                    posts = self.blockchain.rpc.get_replies_by_last_update(
                        reduced_query["start_parent_author"],
                        reduced_query["start_permlink"],
                        reduced_query["limit"],
                    )
            if len(posts) == 0:
                if (
                    "start_parent_author" in reduced_query and reduced_query["start_parent_author"]
                ) and ("start_permlink" in reduced_query and reduced_query["start_permlink"]):
                    if "limit" in reduced_query:
                        posts = self.blockchain.rpc.get_replies_by_last_update(
                            reduced_query["start_parent_author"],
                            reduced_query["start_permlink"],
                            reduced_query["limit"],
                        )
                    else:
                        posts = self.blockchain.rpc.get_replies_by_last_update(
                            reduced_query["start_parent_author"],
                            reduced_query["start_permlink"],
                            100,
                        )
        if posts is None:
            posts = []
        if raw_data:
            super(Discussions_by_replies, self).__init__([x for x in posts])
        else:
            super(Discussions_by_replies, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Replies_by_last_update(list):
    """Returns a list of replies by last update

    :param Query discussion_query: Defines the parameter
        searching posts start_parent_author and start_permlink must be set.
    :param bool use_appbase: use condenser call when set to False, default is False
    :param bool raw_data: returns list of comments when False, default is False
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Replies_by_last_update
        q = Query(limit=10, start_parent_author="steemit", start_permlink="firstpost")
        for h in Replies_by_last_update(q):
            print(h)

    """

    def __init__(
        self,
        discussion_query,
        lazy=False,
        use_appbase=False,
        raw_data=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        posts = []
        if self.blockchain.rpc.get_use_appbase() and use_appbase:
            try:
                posts = self.blockchain.rpc.get_replies_by_last_update(discussion_query, api="tags")
                if "discussions" in posts:
                    posts = posts["discussions"]
            except Exception:
                posts = self.blockchain.rpc.get_replies_by_last_update(
                    discussion_query["start_author"],
                    discussion_query["start_permlink"],
                    discussion_query["limit"],
                )
        if len(posts) == 0:
            posts = self.blockchain.rpc.get_replies_by_last_update(
                discussion_query["start_author"],
                discussion_query["start_permlink"],
                discussion_query["limit"],
            )
        if posts is None:
            posts = []
        if raw_data:
            super(Replies_by_last_update, self).__init__([x for x in posts])
        else:
            super(Replies_by_last_update, self).__init__(
                [Comment(x, lazy=lazy, blockchain_instance=self.blockchain) for x in posts]
            )


class Trending_tags(list):
    """Get trending tags

    :param Query discussion_query: Defines the parameter
        searching posts, start_tag is used if set
    :param Steem blockchain_instance: Steem instance

    .. testcode::

        from nectar.discussions import Query, Trending_tags
        q = Query(limit=10)
        for h in Trending_tags(q):
            print(h)

    """

    def __init__(
        self, discussion_query, lazy=False, use_appbase=False, blockchain_instance=None, **kwargs
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.blockchain.rpc.set_next_node_on_empty_reply(
            self.blockchain.rpc.get_use_appbase() and use_appbase
        )
        limit = discussion_query["limit"] if "limit" in discussion_query else 0
        tags = []
        try:
            # Try to use bridge API for getting trending tags
            # Unfortunately there's no direct bridge API for tags, so we fall back to condenser API
            if self.blockchain.rpc.get_use_appbase() and use_appbase:
                tags = self.blockchain.rpc.get_trending_tags(
                    {"start": "", "limit": limit}, api="condenser"
                )["tags"]
            else:
                tags = self.blockchain.rpc.get_trending_tags("", limit)
        except Exception:
            # If API fails, return empty list
            pass
        super(Trending_tags, self).__init__(tags)



# -*- coding: utf-8 -*-
import binascii
import bisect
import hashlib
import itertools
import os
import re
import unicodedata
from binascii import hexlify, unhexlify

import ecdsa

from .base58 import Base58, doublesha256, ripemd160
from .bip32 import BIP32Key, parse_path
from .dictionary import words as BrainKeyDictionary
from .dictionary import words_bip39 as MnemonicDictionary
from .prefix import Prefix

PBKDF2_ROUNDS = 2048


# From <https://stackoverflow.com/questions/212358/binary-search-bisection-in-python/2233940#2233940>
def binary_search(a, x, lo=0, hi=None):  # can't use a to specify default for hi
    hi = hi if hi is not None else len(a)  # hi defaults to len(a)
    pos = bisect.bisect_left(a, x, lo, hi)  # find insertion position
    return pos if pos != hi and a[pos] == x else -1  # don't walk off the end


class PasswordKey(Prefix):
    """This class derives a private key given the account name, the
    role and a password. It leverages the technology of Brainkeys
    and allows people to have a secure private key by providing a
    passphrase only.
    """

    def __init__(self, account, password, role="active", prefix=None):
        self.set_prefix(prefix)
        self.account = account
        self.role = role
        self.password = password

    def normalize(self, seed):
        """Correct formating with single whitespace syntax and no trailing space"""
        return " ".join(re.compile("[\t\n\v\f\r ]+").split(seed))

    def get_private(self):
        """Derive private key from the account, the role and the password"""
        if self.account is None and self.role is None:
            seed = self.password
        elif self.account == "" and self.role == "":
            seed = self.password
        else:
            seed = self.account + self.role + self.password
        seed = self.normalize(seed)
        a = bytes(seed, "utf8")
        s = hashlib.sha256(a).digest()
        return PrivateKey(hexlify(s).decode("ascii"), prefix=self.prefix)

    def get_public(self):
        return self.get_private().pubkey

    def get_private_key(self):
        return self.get_private()

    def get_public_key(self):
        return self.get_public()


class BrainKey(Prefix):
    """Brainkey implementation similar to the graphene-ui web-wallet.

    :param str brainkey: Brain Key
    :param int sequence: Sequence number for consecutive keys

    Keys in Graphene are derived from a seed brain key which is a string of
    16 words out of a predefined dictionary with 49744 words. It is a
    simple single-chain key derivation scheme that is not compatible with
    BIP44 but easy to use.

    Given the brain key, a private key is derived as::

        privkey = SHA256(SHA512(brainkey + " " + sequence))

    Incrementing the sequence number yields a new key that can be
    regenerated given the brain key.

    """

    def __init__(self, brainkey=None, sequence=0, prefix=None):
        self.set_prefix(prefix)
        if not brainkey:
            self.brainkey = self.suggest()
        else:
            self.brainkey = self.normalize(brainkey).strip()
        self.sequence = sequence

    def __next__(self):
        """Get the next private key (sequence number increment) for
        iterators
        """
        return self.next_sequence()

    def next_sequence(self):
        """Increment the sequence number by 1"""
        self.sequence += 1
        return self

    def normalize(self, brainkey):
        """Correct formating with single whitespace syntax and no trailing space"""
        return " ".join(re.compile("[\t\n\v\f\r ]+").split(brainkey))

    def get_brainkey(self):
        """Return brain key of this instance"""
        return self.normalize(self.brainkey)

    def get_private(self):
        """Derive private key from the brain key and the current sequence
        number
        """
        encoded = "%s %d" % (self.brainkey, self.sequence)
        a = bytes(encoded, "ascii")
        s = hashlib.sha256(hashlib.sha512(a).digest()).digest()
        return PrivateKey(hexlify(s).decode("ascii"), prefix=self.prefix)

    def get_blind_private(self):
        """Derive private key from the brain key (and no sequence number)"""
        a = bytes(self.brainkey, "ascii")
        return PrivateKey(hashlib.sha256(a).hexdigest(), prefix=self.prefix)

    def get_public(self):
        return self.get_private().pubkey

    def get_private_key(self):
        return self.get_private()

    def get_public_key(self):
        return self.get_public()

    def suggest(self, word_count=16):
        """Suggest a new random brain key. Randomness is provided by the
        operating system using ``os.urandom()``.
        """
        brainkey = [None] * word_count
        dict_lines = BrainKeyDictionary.split(",")
        if not len(dict_lines) == 49744:
            raise AssertionError()
        for j in range(0, word_count):
            urand = os.urandom(2)
            num = int.from_bytes(urand, byteorder="little")
            rndMult = num / 2**16  # returns float between 0..1 (inclusive)
            wIdx = int(round(len(dict_lines) * rndMult))
            brainkey[j] = dict_lines[wIdx]
        return " ".join(brainkey).upper()


# From https://github.com/trezor/python-mnemonic/blob/master/mnemonic/mnemonic.py
#
# Copyright (c) 2013 Pavol Rusnak
# Copyright (c) 2017 mruddy
class Mnemonic(object):
    """BIP39 mnemoric implementation"""

    def __init__(self):
        self.wordlist = MnemonicDictionary.split(",")
        self.radix = 2048

    def generate(self, strength=128):
        """Generates a word list based on the given strength

        :param int strength: initial entropy strength, must be one of [128, 160, 192, 224, 256]

        """
        if strength not in [128, 160, 192, 224, 256]:
            raise ValueError(
                "Strength should be one of the following: [128, 160, 192, 224, 256], but it is not (%d)."
                % strength
            )
        return self.to_mnemonic(os.urandom(strength // 8))

    # Adapted from <http://tinyurl.com/oxmn476>
    def to_entropy(self, words):
        if not isinstance(words, list):
            words = words.split(" ")
        if len(words) not in [12, 15, 18, 21, 24]:
            raise ValueError(
                "Number of words must be one of the following: [12, 15, 18, 21, 24], but it is not (%d)."
                % len(words)
            )
        # Look up all the words in the list and construct the
        # concatenation of the original entropy and the checksum.
        concatLenBits = len(words) * 11
        concatBits = [False] * concatLenBits
        wordindex = 0
        use_binary_search = True
        for word in words:
            # Find the words index in the wordlist
            ndx = (
                binary_search(self.wordlist, word)
                if use_binary_search
                else self.wordlist.index(word)
            )
            if ndx < 0:
                raise LookupError('Unable to find "%s" in word list.' % word)
            # Set the next 11 bits to the value of the index.
            for ii in range(11):
                concatBits[(wordindex * 11) + ii] = (ndx & (1 << (10 - ii))) != 0
            wordindex += 1
        checksumLengthBits = concatLenBits // 33
        entropyLengthBits = concatLenBits - checksumLengthBits
        # Extract original entropy as bytes.
        entropy = bytearray(entropyLengthBits // 8)
        for ii in range(len(entropy)):
            for jj in range(8):
                if concatBits[(ii * 8) + jj]:
                    entropy[ii] |= 1 << (7 - jj)
        # Take the digest of the entropy.
        hashBytes = hashlib.sha256(entropy).digest()
        hashBits = list(
            itertools.chain.from_iterable(
                ([c & (1 << (7 - i)) != 0 for i in range(8)] for c in hashBytes)
            )
        )
        # Check all the checksum bits.
        for i in range(checksumLengthBits):
            if concatBits[entropyLengthBits + i] != hashBits[i]:
                raise ValueError("Failed checksum.")
        return entropy

    def to_mnemonic(self, data):
        if len(data) not in [16, 20, 24, 28, 32]:
            raise ValueError(
                "Data length should be one of the following: [16, 20, 24, 28, 32], but it is not (%d)."
                % len(data)
            )
        h = hashlib.sha256(data).hexdigest()
        b = (
            bin(int(binascii.hexlify(data), 16))[2:].zfill(len(data) * 8)
            + bin(int(h, 16))[2:].zfill(256)[: len(data) * 8 // 32]
        )
        result = []
        for i in range(len(b) // 11):
            idx = int(b[i * 11 : (i + 1) * 11], 2)
            result.append(self.wordlist[idx])

        result_phrase = " ".join(result)
        return result_phrase

    def check(self, mnemonic):
        """Checks the mnemonic word list is valid
        :param list mnemonic: mnemonic word list with lenght of 12, 15, 18, 21, 24
        :returns: True, when valid
        """
        mnemonic = self.normalize_string(mnemonic).split(" ")
        # list of valid mnemonic lengths
        if len(mnemonic) not in [12, 15, 18, 21, 24]:
            return False
        try:
            idx = map(lambda x: bin(self.wordlist.index(x))[2:].zfill(11), mnemonic)
            b = "".join(idx)
        except ValueError:
            return False
        l = len(b)  # noqa: E741
        d = b[: l // 33 * 32]
        h = b[-l // 33 :]
        nd = binascii.unhexlify(hex(int(d, 2))[2:].rstrip("L").zfill(l // 33 * 8))
        nh = bin(int(hashlib.sha256(nd).hexdigest(), 16))[2:].zfill(256)[: l // 33]
        return h == nh

    def check_word(self, word):
        return word in self.wordlist

    def expand_word(self, prefix):
        """Expands a word when sufficient chars are given

        :param str prefix: first chars of a valid dict word

        """
        if prefix in self.wordlist:
            return prefix
        else:
            matches = [word for word in self.wordlist if word.startswith(prefix)]
            if len(matches) == 1:  # matched exactly one word in the wordlist
                return matches[0]
            else:
                # exact match not found.
                # this is not a validation routine, just return the input
                return prefix

    def expand(self, mnemonic):
        """Expands all words given in a list"""
        return " ".join(map(self.expand_word, mnemonic.split(" ")))

    @classmethod
    def normalize_string(cls, txt):
        """Normalizes strings"""
        utxt = txt.decode("utf8")
        return unicodedata.normalize("NFKD", utxt)

    @classmethod
    def to_seed(cls, mnemonic, passphrase=""):
        """Returns a seed based on bip39

        :param str mnemonic: string containing a valid mnemonic word list
        :param str passphrase: optional, passphrase can be set to modify the returned seed.

        """
        mnemonic = cls.normalize_string(mnemonic)
        passphrase = cls.normalize_string(passphrase)
        passphrase = "mnemonic" + passphrase
        mnemonic = mnemonic.encode("utf-8")
        passphrase = passphrase.encode("utf-8")
        stretched = hashlib.pbkdf2_hmac("sha512", mnemonic, passphrase, PBKDF2_ROUNDS)
        return stretched[:64]


class MnemonicKey(Prefix):
    """This class derives a private key from a BIP39 mnemoric implementation"""

    def __init__(
        self, word_list=None, passphrase="", account_sequence=0, key_sequence=0, prefix=None
    ):
        self.set_prefix(prefix)
        if word_list is not None:
            self.set_mnemonic(word_list, passphrase=passphrase)
        else:
            self.seed = None
        self.account_sequence = account_sequence
        self.key_sequence = key_sequence
        self.prefix = prefix
        self.path = "m/48'/13'/0'/%d'/%d'" % (self.account_sequence, self.key_sequence)

    def set_mnemonic(self, word_list, passphrase=""):
        mnemonic = Mnemonic()
        if not mnemonic.check(word_list):
            raise ValueError("Word list is not valid!")
        self.seed = mnemonic.to_seed(word_list, passphrase=passphrase)

    def generate_mnemonic(self, passphrase="", strength=256):
        mnemonic = Mnemonic()
        word_list = mnemonic.generate(strength=strength)
        self.seed = mnemonic.to_seed(word_list, passphrase=passphrase)
        return word_list

    def set_path_BIP32(self, path):
        self.path = path

    def set_path_BIP44(
        self, account_sequence=0, chain_sequence=0, key_sequence=0, hardened_address=True
    ):
        if account_sequence < 0:
            raise ValueError("account_sequence must be >= 0")
        if key_sequence < 0:
            raise ValueError("key_sequence must be >= 0")
        if chain_sequence < 0:
            raise ValueError("chain_sequence must be >= 0")
        self.account_sequence = account_sequence
        self.key_sequence = key_sequence
        if hardened_address:
            self.path = "m/44'/0'/%d'/%d/%d'" % (
                self.account_sequence,
                chain_sequence,
                self.key_sequence,
            )
        else:
            self.path = "m/44'/0'/%d'/%d/%d" % (
                self.account_sequence,
                chain_sequence,
                self.key_sequence,
            )

    def set_path_BIP48(self, network_index=13, role="owner", account_sequence=0, key_sequence=0):
        if account_sequence < 0:
            raise ValueError("account_sequence must be >= 0")
        if key_sequence < 0:
            raise ValueError("key_sequence must be >= 0")
        if network_index < 0:
            raise ValueError("network_index must be >= 0")
        if isinstance(role, str) and role not in ["owner", "active", "posting", "memo"]:
            raise ValueError("Wrong role!")
        elif isinstance(role, int) and role < 0:
            raise ValueError("role must be >= 0")
        if role == "owner":
            role = 0
        elif role == "active":
            role = 1
        elif role == "posting":
            role = 4
        elif role == "memo":
            role = 3

        self.account_sequence = account_sequence
        self.key_sequence = key_sequence
        self.path = "m/48'/%d'/%d'/%d'/%d'" % (
            network_index,
            role,
            self.account_sequence,
            self.key_sequence,
        )

    def next_account_sequence(self):
        """Increment the account sequence number by 1"""
        self.account_sequence += 1
        return self

    def next_sequence(self):
        """Increment the key sequence number by 1"""
        self.key_sequence += 1
        return self

    def set_path(self, path):
        self.path = path

    def get_path(self):
        return self.path

    def get_private(self):
        """Derive private key from the account_sequence, the role and the key_sequence"""
        if self.seed is None:
            raise ValueError("seed is None, set or generate a mnemnoric first")
        key = BIP32Key.fromEntropy(self.seed)
        for n in parse_path(self.get_path()):
            key = key.ChildKey(n)
        return PrivateKey(key.WalletImportFormat(), prefix=self.prefix)

    def get_public(self):
        return self.get_private().pubkey

    def get_private_key(self):
        return self.get_private()

    def get_public_key(self):
        return self.get_public()


class Address(Prefix):
    """Address class

    This class serves as an address representation for Public Keys.

    :param str address: Base58 encoded address (defaults to ``None``)
    :param str prefix: Network prefix (defaults to ``STM``)

    Example::

       Address("STMFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi")

    """

    def __init__(self, address, prefix=None):
        self.set_prefix(prefix)
        self._address = Base58(address, prefix=self.prefix)

    @classmethod
    def from_pubkey(cls, pubkey, compressed=True, version=56, prefix=None):
        """Load an address provided by the public key.
        Version: 56 => PTS
        """
        # Ensure this is a public key
        pubkey = PublicKey(pubkey, prefix=prefix or Prefix.prefix)
        if compressed:
            pubkey_plain = pubkey.compressed()
        else:
            pubkey_plain = pubkey.uncompressed()
        sha = hashlib.sha256(unhexlify(pubkey_plain)).hexdigest()
        rep = hexlify(ripemd160(sha)).decode("ascii")
        s = ("%.2x" % version) + rep
        result = s + hexlify(doublesha256(s)[:4]).decode("ascii")
        result = hexlify(ripemd160(result)).decode("ascii")
        return cls(result, prefix=pubkey.prefix)

    @classmethod
    def derivesha256address(cls, pubkey, compressed=True, prefix=None):
        """Derive address using ``RIPEMD160(SHA256(x))``"""
        pubkey = PublicKey(pubkey, prefix=prefix or Prefix.prefix)
        if compressed:
            pubkey_plain = pubkey.compressed()
        else:
            pubkey_plain = pubkey.uncompressed()
        pkbin = unhexlify(repr(pubkey_plain))
        result = hexlify(hashlib.sha256(pkbin).digest())
        result = hexlify(ripemd160(result)).decode("ascii")
        return cls(result, prefix=pubkey.prefix)

    @classmethod
    def derivesha512address(cls, pubkey, compressed=True, prefix=None):
        """Derive address using ``RIPEMD160(SHA512(x))``"""
        pubkey = PublicKey(pubkey, prefix=prefix or Prefix.prefix)
        if compressed:
            pubkey_plain = pubkey.compressed()
        else:
            pubkey_plain = pubkey.uncompressed()
        pkbin = unhexlify(repr(pubkey_plain))
        result = hexlify(hashlib.sha512(pkbin).digest())
        result = hexlify(ripemd160(result)).decode("ascii")
        return cls(result, prefix=pubkey.prefix)

    def __repr__(self):
        """Gives the hex representation of the ``GrapheneBase58CheckEncoded``
        Graphene address.
        """
        return repr(self._address)

    def __str__(self):
        """Returns the readable Graphene address. This call is equivalent to
        ``format(Address, "STM")``
        """
        return format(self._address, self.prefix)

    def __format__(self, _format):
        """May be issued to get valid "MUSE", "PLAY" or any other Graphene compatible
        address with corresponding prefix.
        """
        return format(self._address, _format)

    def __bytes__(self):
        """Returns the raw content of the ``Base58CheckEncoded`` address"""
        return bytes(self._address)


class GrapheneAddress(Address):
    """Graphene Addresses are different. Hence we have a different class"""

    @classmethod
    def from_pubkey(cls, pubkey, compressed=True, version=56, prefix=None):
        # Ensure this is a public key
        pubkey = PublicKey(pubkey, prefix=prefix or Prefix.prefix)
        if compressed:
            pubkey_plain = pubkey.compressed()
        else:
            pubkey_plain = pubkey.uncompressed()

        """ Derive address using ``RIPEMD160(SHA512(x))`` """
        addressbin = ripemd160(hashlib.sha512(unhexlify(pubkey_plain)).hexdigest())
        result = Base58(hexlify(addressbin).decode("ascii"))
        return cls(result, prefix=pubkey.prefix)


class PublicKey(Prefix):
    """This class deals with Public Keys and inherits ``Address``.

    :param str pk: Base58 encoded public key
    :param str prefix: Network prefix (defaults to ``STM``)

    Example::

       PublicKey("STM6UtYWWs3rkZGV8JA86qrgkG6tyFksgECefKE1MiH4HkLD8PFGL")

    .. note:: By default, graphene-based networks deal with **compressed**
              public keys. If an **uncompressed** key is required, the
              method :func:`unCompressed` can be used::

                  PublicKey("xxxxx").unCompressed()

    """

    def __init__(self, pk, prefix=None):
        """Init PublicKey
        :param str pk: Base58 encoded public key
        :param str prefix: Network prefix (defaults to ``STM``)
        """
        self.set_prefix(prefix)
        if isinstance(pk, PublicKey):
            pk = format(pk, self.prefix)

        if str(pk).startswith("04"):
            # We only ever deal with compressed keys, so let's make it
            # compressed
            order = ecdsa.SECP256k1.order
            p = ecdsa.VerifyingKey.from_string(
                unhexlify(pk[2:]), curve=ecdsa.SECP256k1
            ).pubkey.point
            x_str = ecdsa.util.number_to_string(p.x(), order)
            pk = hexlify(chr(2 + (p.y() & 1)).encode("ascii") + x_str).decode("ascii")

        self._pk = Base58(pk, prefix=self.prefix)

    @property
    def pubkey(self):
        return self._pk

    def get_public_key(self):
        """Returns the pubkey"""
        return self.pubkey

    @property
    def compressed_key(self):
        return PublicKey(self.compressed())

    def _derive_y_from_x(self, x, is_even):
        """Derive y point from x point"""
        curve = ecdsa.SECP256k1.curve
        # The curve equation over F_p is:
        #   y^2 = x^3 + ax + b
        a, b, p = curve.a(), curve.b(), curve.p()
        alpha = (pow(x, 3, p) + a * x + b) % p
        beta = ecdsa.numbertheory.square_root_mod_prime(alpha, p)
        if (beta % 2) == is_even:
            beta = p - beta
        return beta

    def compressed(self):
        """Derive compressed public key"""
        return repr(self._pk)

    def uncompressed(self):
        """Derive uncompressed key"""
        public_key = repr(self._pk)
        prefix = public_key[0:2]
        if prefix == "04":
            return public_key
        if not (prefix == "02" or prefix == "03"):
            raise AssertionError()
        x = int(public_key[2:], 16)
        y = self._derive_y_from_x(x, (prefix == "02"))
        key = "04" + "%064x" % x + "%064x" % y
        return key

    def point(self):
        """Return the point for the public key"""
        string = unhexlify(self.uncompressed())
        return ecdsa.VerifyingKey.from_string(string[1:], curve=ecdsa.SECP256k1).pubkey.point

    def child(self, offset256):
        """Derive new public key from this key and a sha256 "offset" """
        a = bytes(self) + offset256
        s = hashlib.sha256(a).digest()
        return self.add(s)

    def add(self, digest256):
        """Derive new public key from this key and a sha256 "digest" """
        from .ecdsa import tweakaddPubkey

        return tweakaddPubkey(self, digest256)

    @classmethod
    def from_privkey(cls, privkey, prefix=None):
        """Derive uncompressed public key"""
        privkey = PrivateKey(privkey, prefix=prefix or Prefix.prefix)
        secret = unhexlify(repr(privkey))
        order = ecdsa.SigningKey.from_string(secret, curve=ecdsa.SECP256k1).curve.generator.order()
        p = ecdsa.SigningKey.from_string(secret, curve=ecdsa.SECP256k1).verifying_key.pubkey.point
        x_str = ecdsa.util.number_to_string(p.x(), order)
        # y_str = ecdsa.util.number_to_string(p.y(), order)
        compressed = hexlify(chr(2 + (p.y() & 1)).encode("ascii") + x_str).decode("ascii")
        # uncompressed = hexlify(
        #    chr(4).encode('ascii') + x_str + y_str).decode('ascii')
        return cls(compressed, prefix=prefix or Prefix.prefix)

    def __repr__(self):
        """Gives the hex representation of the Graphene public key."""
        return repr(self._pk)

    def __str__(self):
        """Returns the readable Graphene public key. This call is equivalent to
        ``format(PublicKey, "STM")``
        """
        return format(self._pk, self.prefix)

    def __format__(self, _format):
        """Formats the instance of:doc:`Base58 <base58>` according to ``_format``"""
        return format(self._pk, _format)

    def __bytes__(self):
        """Returns the raw public key (has length 33)"""
        return bytes(self._pk)

    def __lt__(self, other):
        """For sorting of public keys (due to graphene),
        we actually sort according to addresses
        """
        assert isinstance(other, PublicKey)
        return repr(self.address) < repr(other.address)

    def unCompressed(self):
        """Alias for self.uncompressed() - LEGACY"""
        return self.uncompressed()

    @property
    def address(self):
        """Obtain a GrapheneAddress from a public key"""
        return GrapheneAddress.from_pubkey(repr(self), prefix=self.prefix)


class PrivateKey(Prefix):
    """Derives the compressed and uncompressed public keys and
    constructs two instances of :class:`PublicKey`:

    :param str wif: Base58check-encoded wif key
    :param str prefix: Network prefix (defaults to ``STM``)

    Example::

        PrivateKey("5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd")

    Compressed vs. Uncompressed:

    * ``PrivateKey("w-i-f").pubkey``:
        Instance of :class:`PublicKey` using compressed key.
    * ``PrivateKey("w-i-f").pubkey.address``:
        Instance of :class:`Address` using compressed key.
    * ``PrivateKey("w-i-f").uncompressed``:
        Instance of :class:`PublicKey` using uncompressed key.
    * ``PrivateKey("w-i-f").uncompressed.address``:
        Instance of :class:`Address` using uncompressed key.

    """

    def __init__(self, wif=None, prefix=None):
        self.set_prefix(prefix)
        if wif is None:
            import os

            self._wif = Base58(hexlify(os.urandom(32)).decode("ascii"))
        elif isinstance(wif, PrivateKey):
            self._wif = wif._wif
        elif isinstance(wif, Base58):
            self._wif = wif
        else:
            self._wif = Base58(wif)

        assert len(repr(self._wif)) == 64

    @property
    def bitcoin(self):
        return BitcoinPublicKey.from_privkey(self)

    @property
    def address(self):
        return Address.from_pubkey(self.pubkey, prefix=self.prefix)

    @property
    def pubkey(self):
        return self.compressed

    def get_public_key(self):
        """Legacy: Returns the pubkey"""
        return self.pubkey

    @property
    def compressed(self):
        return PublicKey.from_privkey(self, prefix=self.prefix)

    @property
    def uncompressed(self):
        return PublicKey(self.pubkey.uncompressed(), prefix=self.prefix)

    def get_secret(self):
        """Get sha256 digest of the wif key."""
        return hashlib.sha256(bytes(self)).digest()

    def derive_private_key(self, sequence):
        """Derive new private key from this private key and an arbitrary
        sequence number
        """
        encoded = "%s %d" % (str(self), sequence)
        a = bytes(encoded, "ascii")
        s = hashlib.sha256(hashlib.sha512(a).digest()).digest()
        return PrivateKey(hexlify(s).decode("ascii"), prefix=self.pubkey.prefix)

    def child(self, offset256):
        """Derive new private key from this key and a sha256 "offset" """
        a = bytes(self.pubkey) + offset256
        s = hashlib.sha256(a).digest()
        return self.derive_from_seed(s)

    def derive_from_seed(self, offset):
        """Derive private key using "generate_from_seed" method.
        Here, the key itself serves as a `seed`, and `offset`
        is expected to be a sha256 digest.
        """
        seed = int(hexlify(bytes(self)).decode("ascii"), 16)
        z = int(hexlify(offset).decode("ascii"), 16)
        order = ecdsa.SECP256k1.order
        secexp = (seed + z) % order
        secret = "%0x" % secexp
        if len(secret) < 64:  # left-pad with zeroes
            secret = ("0" * (64 - len(secret))) + secret
        return PrivateKey(secret, prefix=self.pubkey.prefix)

    def __format__(self, _format):
        """Formats the instance of:doc:`Base58 <base58>` according to
        ``_format``
        """
        return format(self._wif, _format)

    def __repr__(self):
        """Gives the hex representation of the Graphene private key."""
        return repr(self._wif)

    def __str__(self):
        """Returns the readable (uncompressed wif format) Graphene private key. This
        call is equivalent to ``format(PrivateKey, "WIF")``
        """
        return format(self._wif, "WIF")

    def __bytes__(self):
        """Returns the raw private key"""
        return bytes(self._wif)


class BitcoinAddress(Address):
    @classmethod
    def from_pubkey(cls, pubkey, compressed=False, version=56, prefix=None):
        # Ensure this is a public key
        pubkey = PublicKey(pubkey)
        if compressed:
            pubkey = pubkey.compressed()
        else:
            pubkey = pubkey.uncompressed()

        """ Derive address using ``RIPEMD160(SHA256(x))`` """
        addressbin = ripemd160(hexlify(hashlib.sha256(unhexlify(pubkey)).digest()))
        return cls(hexlify(addressbin).decode("ascii"))

    def __str__(self):
        """Returns the readable Graphene address. This call is equivalent to
        ``format(Address, "GPH")``
        """
        return format(self._address, "BTC")


class BitcoinPublicKey(PublicKey):
    @property
    def address(self):
        return BitcoinAddress.from_pubkey(repr(self))



# -*- coding: utf-8 -*-
import logging
import math
from datetime import date, datetime, timezone

from nectar.blockchaininstance import BlockChainInstance
from nectar.constants import STEEM_100_PERCENT
from nectargraphenebase.chains import known_chains

from .amount import Amount
from .utils import formatToTimeStamp

log = logging.getLogger(__name__)


class Hive(BlockChainInstance):
    """Connect to the Hive network.

    :param str node: Node to connect to *(optional)*
    :param str rpcuser: RPC user *(optional)*
    :param str rpcpassword: RPC password *(optional)*
    :param bool nobroadcast: Do **not** broadcast a transaction!
        *(optional)*
    :param bool unsigned: Do **not** sign a transaction! *(optional)*
    :param bool debug: Enable Debugging *(optional)*
    :param keys: Predefine the wif keys to shortcut the
        wallet database *(optional)*
    :type keys: array, dict, string
    :param wif: Predefine the wif keys to shortcut the
            wallet database *(optional)*
    :type wif: array, dict, string
    :param bool offline: Boolean to prevent connecting to network (defaults
        to ``False``) *(optional)*
    :param int expiration: Delay in seconds until transactions are supposed
        to expire *(optional)* (default is 30)
    :param str blocking: Wait for broadcasted transactions to be included
        in a block and return full transaction (can be "head" or
        "irreversible")
    :param bool bundle: Do not broadcast transactions right away, but allow
        to bundle operations. It is not possible to send out more than one
        vote operation and more than one comment operation in a single broadcast *(optional)*
    :param bool appbase: Use the new appbase rpc protocol on nodes with version
        0.19.4 or higher. The settings has no effect on nodes with version of 0.19.3 or lower.
    :param int num_retries: Set the maximum number of reconnects to the nodes before
        NumRetriesReached is raised. Disabled for -1. (default is -1)
    :param int num_retries_call: Repeat num_retries_call times a rpc call on node error (default is 5)
    :param int timeout: Timeout setting for https nodes (default is 60)
    :param bool use_hs: When True, a hivesigner object is created. Can be used for
        broadcast posting op or creating hot_links (default is False)
    :param HiveSigner hivesigner: A HiveSigner object can be set manually, set use_hs to True
    :param dict custom_chains: custom chain which should be added to the known chains

    Three wallet operation modes are possible:

    * **Wallet Database**: Here, the nectarlibs load the keys from the
      locally stored wallet SQLite database (see ``storage.py``).
      To use this mode, simply call ``Hive()`` without the
      ``keys`` parameter
    * **Providing Keys**: Here, you can provide the keys for
      your accounts manually. All you need to do is add the wif
      keys for the accounts you want to use as a simple array
      using the ``keys`` parameter to ``Steem()``.
    * **Force keys**: This more is for advanced users and
      requires that you know what you are doing. Here, the
      ``keys`` parameter is a dictionary that overwrite the
      ``active``, ``owner``, ``posting`` or ``memo`` keys for
      any account. This mode is only used for *foreign*
      signatures!

    If no node is provided, it will connect to default nodes from
    nectar.NodeList. Default settings can be changed with:

    .. code-block:: python

        hive = Hive(<host>)

    where ``<host>`` starts with ``https://``, ``ws://`` or ``wss://``.

    The purpose of this class it to simplify interaction with
    Hive.

    The idea is to have a class that allows to do this:

    .. code-block:: python

        >>> from nectar import Hive
        >>> hive = Hive()
        >>> print(hive.get_blockchain_version())  # doctest: +SKIP

    This class also deals with edits, votes and reading content.

    Example for adding a custom chain:

    .. code-block:: python

        from nectar import Hive
        stm = Hive(node=["https://mytstnet.com"], custom_chains={"MYTESTNET":
            {'chain_assets': [{'asset': 'HBD', 'id': 0, 'precision': 3, 'symbol': 'HBD'},
                              {'asset': 'STEEM', 'id': 1, 'precision': 3, 'symbol': 'STEEM'},
                              {'asset': 'VESTS', 'id': 2, 'precision': 6, 'symbol': 'VESTS'}],
             'chain_id': '79276aea5d4877d9a25892eaa01b0adf019d3e5cb12a97478df3298ccdd01674',
             'min_version': '0.0.0',
             'prefix': 'MTN'}
            }
        )

    """

    def get_network(self, use_stored_data=True, config=None):
        """Identify the network

        :param bool use_stored_data: if True, stored data will be returned. If stored data are
                                     empty or old, refresh_data() is used.

        :returns: Network parameters
        :rtype: dictionary
        """
        if use_stored_data:
            self.refresh_data("config")
            return self.data["network"]

        if self.rpc is None:
            return known_chains["HIVE"]
        try:
            return self.rpc.get_network(props=config)
        except Exception:
            return known_chains["HIVE"]

    def rshares_to_token_backed_dollar(
        self, rshares, not_broadcasted_vote=False, use_stored_data=True
    ):
        return self.rshares_to_hbd(
            rshares, not_broadcasted_vote=not_broadcasted_vote, use_stored_data=use_stored_data
        )

    def rshares_to_hbd(self, rshares, not_broadcasted_vote=False, use_stored_data=True):
        """Calculates the current HBD value of a vote"""
        payout = float(rshares) * self.get_hbd_per_rshares(
            use_stored_data=use_stored_data,
            not_broadcasted_vote_rshares=rshares if not_broadcasted_vote else 0,
        )
        return payout

    def get_hbd_per_rshares(self, not_broadcasted_vote_rshares=0, use_stored_data=True):
        """Returns the current rshares to HBD ratio"""
        reward_fund = self.get_reward_funds(use_stored_data=use_stored_data)
        reward_balance = float(Amount(reward_fund["reward_balance"], blockchain_instance=self))
        recent_claims = float(reward_fund["recent_claims"]) + not_broadcasted_vote_rshares
        if recent_claims == 0:
            return 0
        fund_per_share = reward_balance / (recent_claims)
        median_price = self.get_median_price(use_stored_data=use_stored_data)
        if median_price is None:
            return 0
        HBD_price = float(median_price * (Amount(1, self.hive_symbol, blockchain_instance=self)))
        return fund_per_share * HBD_price

    def get_token_per_mvest(self, time_stamp=None, use_stored_data=True):
        return self.get_hive_per_mvest(time_stamp=time_stamp, use_stored_data=use_stored_data)

    def get_hive_per_mvest(self, time_stamp=None, use_stored_data=True):
        """Returns the MVEST to HIVE ratio

        :param int time_stamp: (optional) if set, return an estimated
            HIVE per MVEST ratio for the given time stamp. If unset the
            current ratio is returned (default). (can also be a datetime object)
        """
        if self.offline and time_stamp is None:
            time_stamp = datetime.now(timezone.utc)

        if time_stamp is not None:
            if isinstance(time_stamp, (datetime, date)):
                time_stamp = formatToTimeStamp(time_stamp)
            a = 2.1325476281078992e-05
            b = -31099.685481490847
            a2 = 2.9019227739473682e-07
            b2 = 48.41432402074669

            if time_stamp < (b2 - b) / (a - a2):
                return a * time_stamp + b
            else:
                return a2 * time_stamp + b2
        global_properties = self.get_dynamic_global_properties(use_stored_data=use_stored_data)
        return float(
            Amount(global_properties["total_vesting_fund_hive"], blockchain_instance=self)
        ) / (
            float(Amount(global_properties["total_vesting_shares"], blockchain_instance=self)) / 1e6
        )

    def vests_to_hp(self, vests, timestamp=None, use_stored_data=True):
        """Converts vests to HP

        :param amount.Amount vests/float vests: Vests to convert
        :param int timestamp: (Optional) Can be used to calculate
            the conversion rate from the past

        """
        if isinstance(vests, Amount):
            vests = float(vests)
        return (
            float(vests) / 1e6 * self.get_hive_per_mvest(timestamp, use_stored_data=use_stored_data)
        )

    def vests_to_token_power(self, vests, timestamp=None, use_stored_data=True):
        return self.vests_to_hp(vests, timestamp=timestamp, use_stored_data=use_stored_data)

    def hp_to_vests(self, hp, timestamp=None, use_stored_data=True):
        """Converts HP to vests

        :param float hp: Hive power to convert
        :param datetime timestamp: (Optional) Can be used to calculate
            the conversion rate from the past
        """
        return hp * 1e6 / self.get_hive_per_mvest(timestamp, use_stored_data=use_stored_data)

    def token_power_to_vests(self, token_power, timestamp=None, use_stored_data=True):
        return self.hp_to_vests(token_power, timestamp=timestamp, use_stored_data=use_stored_data)

    def token_power_to_token_backed_dollar(
        self,
        token_power,
        post_rshares=0,
        voting_power=STEEM_100_PERCENT,
        vote_pct=STEEM_100_PERCENT,
        not_broadcasted_vote=True,
        use_stored_data=True,
    ):
        return self.hp_to_hbd(
            token_power,
            post_rshares=post_rshares,
            voting_power=voting_power,
            vote_pct=vote_pct,
            not_broadcasted_vote=not_broadcasted_vote,
            use_stored_data=use_stored_data,
        )

    def hp_to_hbd(
        self,
        hp,
        post_rshares=0,
        voting_power=STEEM_100_PERCENT,
        vote_pct=STEEM_100_PERCENT,
        not_broadcasted_vote=True,
        use_stored_data=True,
    ):
        """Obtain the resulting HBD vote value from Hive power

        :param number hive_power: Hive Power
        :param int post_rshares: rshares of post which is voted
        :param int voting_power: voting power (100% = 10000)
        :param int vote_pct: voting percentage (100% = 10000)
        :param bool not_broadcasted_vote: not_broadcasted or already broadcasted vote (True = not_broadcasted vote).

        Only impactful for very big votes. Slight modification to the value calculation, as the not_broadcasted
        vote rshares decreases the reward pool.
        """
        vesting_shares = int(self.hp_to_vests(hp, use_stored_data=use_stored_data))
        return self.vests_to_hbd(
            vesting_shares,
            post_rshares=post_rshares,
            voting_power=voting_power,
            vote_pct=vote_pct,
            not_broadcasted_vote=not_broadcasted_vote,
            use_stored_data=use_stored_data,
        )

    def vests_to_hbd(
        self,
        vests,
        post_rshares=0,
        voting_power=STEEM_100_PERCENT,
        vote_pct=STEEM_100_PERCENT,
        not_broadcasted_vote=True,
        use_stored_data=True,
    ):
        """Obtain the resulting HBD vote value from vests

        :param number vests: vesting shares
        :param int post_rshares: rshares of post which is voted
        :param int voting_power: voting power (100% = 10000)
        :param int vote_pct: voting percentage (100% = 10000)
        :param bool not_broadcasted_vote: not_broadcasted or already broadcasted vote (True = not_broadcasted vote).

        Only impactful for very big votes. Slight modification to the value calculation, as the not_broadcasted
        vote rshares decreases the reward pool.
        """
        vote_rshares = self.vests_to_rshares(
            vests, post_rshares=post_rshares, voting_power=voting_power, vote_pct=vote_pct
        )
        return self.rshares_to_hbd(
            vote_rshares, not_broadcasted_vote=not_broadcasted_vote, use_stored_data=use_stored_data
        )

    def hp_to_rshares(
        self,
        hive_power,
        post_rshares=0,
        voting_power=STEEM_100_PERCENT,
        vote_pct=STEEM_100_PERCENT,
        use_stored_data=True,
    ):
        """Obtain the r-shares from Hive power

        :param number hive_power: Hive Power
        :param int post_rshares: rshares of post which is voted
        :param int voting_power: voting power (100% = 10000)
        :param int vote_pct: voting percentage (100% = 10000)

        """
        # calculate our account voting shares (from vests)
        vesting_shares = int(self.hp_to_vests(hive_power, use_stored_data=use_stored_data))
        rshares = self.vests_to_rshares(
            vesting_shares,
            post_rshares=post_rshares,
            voting_power=voting_power,
            vote_pct=vote_pct,
            use_stored_data=use_stored_data,
        )
        return rshares

    def vests_to_rshares(
        self,
        vests,
        post_rshares=0,
        voting_power=STEEM_100_PERCENT,
        vote_pct=STEEM_100_PERCENT,
        subtract_dust_threshold=True,
        use_stored_data=True,
    ):
        """Obtain the r-shares from vests

        :param number vests: vesting shares
        :param int post_rshares: rshares of post which is voted
        :param int voting_power: voting power (100% = 10000)
        :param int vote_pct: voting percentage (100% = 10000)

        """
        used_power = self._calc_resulting_vote(
            voting_power=voting_power, vote_pct=vote_pct, use_stored_data=use_stored_data
        )
        # calculate vote rshares
        rshares = int(math.copysign(vests * 1e6 * used_power / STEEM_100_PERCENT, vote_pct))
        if subtract_dust_threshold:
            if abs(rshares) <= self.get_dust_threshold(use_stored_data=use_stored_data):
                return 0
            rshares -= math.copysign(
                self.get_dust_threshold(use_stored_data=use_stored_data), vote_pct
            )
        rshares = self._calc_vote_claim(rshares, post_rshares)
        return rshares

    def hbd_to_rshares(self, hbd, not_broadcasted_vote=False, use_stored_data=True):
        """Obtain the r-shares from HBD

        :param hbd: HBD
        :type hbd: str, int, amount.Amount
        :param bool not_broadcasted_vote: not_broadcasted or already broadcasted vote (True = not_broadcasted vote).
         Only impactful for very high amounts of HBD. Slight modification to the value calculation, as the not_broadcasted
         vote rshares decreases the reward pool.

        """
        if isinstance(hbd, Amount):
            hbd = Amount(hbd, blockchain_instance=self)
        elif isinstance(hbd, str):
            hbd = Amount(hbd, blockchain_instance=self)
        else:
            hbd = Amount(hbd, self.hbd_symbol, blockchain_instance=self)
        if hbd["symbol"] != self.hbd_symbol:
            raise AssertionError("Should input HBD, not any other asset!")

        # If the vote was already broadcasted we can assume the blockchain values to be true
        if not not_broadcasted_vote:
            return int(float(hbd) / self.get_hbd_per_rshares(use_stored_data=use_stored_data))

        # If the vote wasn't broadcasted (yet), we have to calculate the rshares while considering
        # the change our vote is causing to the recent_claims. This is more important for really
        # big votes which have a significant impact on the recent_claims.
        reward_fund = self.get_reward_funds(use_stored_data=use_stored_data)
        median_price = self.get_median_price(use_stored_data=use_stored_data)
        recent_claims = int(reward_fund["recent_claims"])
        reward_balance = Amount(reward_fund["reward_balance"], blockchain_instance=self)
        reward_pool_hbd = median_price * reward_balance
        if hbd > reward_pool_hbd:
            raise ValueError("Provided more HBD than available in the reward pool.")

        # This is the formula we can use to determine the "true" rshares.
        # We get this formula by some math magic using the previous used formulas
        # FundsPerShare = (balance / (claims + newShares)) * Price
        # newShares = amount / FundsPerShare
        # We can now resolve both formulas for FundsPerShare and set the formulas to be equal
        # (balance / (claims + newShares)) * price = amount / newShares
        # Now we resolve for newShares resulting in:
        # newShares = claims * amount / (balance * price - amount)
        rshares = (
            recent_claims
            * float(hbd)
            / ((float(reward_balance) * float(median_price)) - float(hbd))
        )
        return int(rshares)

    def rshares_to_vote_pct(
        self,
        rshares,
        post_rshares=0,
        hive_power=None,
        vests=None,
        voting_power=STEEM_100_PERCENT,
        use_stored_data=True,
    ):
        """Obtain the voting percentage for a desired rshares value
        for a given Hive Power or vesting shares and voting_power
        Give either hive_power or vests, not both.
        When the output is greater than 10000 or less than -10000,
        the given absolute rshares are too high

        Returns the required voting percentage (100% = 10000)

        :param number rshares: desired rshares value
        :param number hive_power: Hive Power
        :param number vests: vesting shares
        :param int voting_power: voting power (100% = 10000)

        """
        if hive_power is None and vests is None:
            raise ValueError("Either hive_power or vests has to be set!")
        if hive_power is not None and vests is not None:
            raise ValueError("Either hive_power or vests has to be set. Not both!")
        if hive_power is not None:
            vests = int(self.hp_to_vests(hive_power, use_stored_data=use_stored_data) * 1e6)

        if self.hardfork >= 20:
            rshares += math.copysign(
                self.get_dust_threshold(use_stored_data=use_stored_data), rshares
            )

        if post_rshares >= 0 and rshares > 0:
            rshares = math.copysign(
                self._calc_revert_vote_claim(abs(rshares), post_rshares), rshares
            )
        elif post_rshares < 0 and rshares < 0:
            rshares = math.copysign(
                self._calc_revert_vote_claim(abs(rshares), abs(post_rshares)), rshares
            )
        elif post_rshares < 0 and rshares > 0:
            rshares = math.copysign(self._calc_revert_vote_claim(abs(rshares), 0), rshares)
        elif post_rshares > 0 and rshares < 0:
            rshares = math.copysign(
                self._calc_revert_vote_claim(abs(rshares), post_rshares), rshares
            )

        max_vote_denom = self._max_vote_denom(use_stored_data=use_stored_data)

        used_power = int(math.ceil(abs(rshares) * STEEM_100_PERCENT / vests))
        used_power = used_power * max_vote_denom

        vote_pct = used_power * STEEM_100_PERCENT / (60 * 60 * 24) / voting_power
        return int(math.copysign(vote_pct, rshares))

    def hbd_to_vote_pct(
        self,
        hbd,
        post_rshares=0,
        hive_power=None,
        vests=None,
        voting_power=STEEM_100_PERCENT,
        not_broadcasted_vote=True,
        use_stored_data=True,
    ):
        """Obtain the voting percentage for a desired HBD value
        for a given Hive Power or vesting shares and voting power
        Give either Hive Power or vests, not both.
        When the output is greater than 10000 or smaller than -10000,
        the HBD value is too high.

        Returns the required voting percentage (100% = 10000)

        :param hbd: desired HBD value
        :type hbd: str, int, amount.Amount
        :param number hive_power: Hive Power
        :param number vests: vesting shares
        :param bool not_broadcasted_vote: not_broadcasted or already broadcasted vote (True = not_broadcasted vote).
         Only impactful for very high amounts of HBD. Slight modification to the value calculation, as the not_broadcasted
         vote rshares decreases the reward pool.

        """
        if isinstance(hbd, Amount):
            hbd = Amount(hbd, blockchain_instance=self)
        elif isinstance(hbd, str):
            hbd = Amount(hbd, blockchain_instance=self)
        else:
            hbd = Amount(hbd, self.hbd_symbol, blockchain_instance=self)
        if hbd["symbol"] != self.hbd_symbol:
            raise AssertionError()
        rshares = self.hbd_to_rshares(
            hbd, not_broadcasted_vote=not_broadcasted_vote, use_stored_data=use_stored_data
        )
        return self.rshares_to_vote_pct(
            rshares,
            post_rshares=post_rshares,
            hive_power=hive_power,
            vests=vests,
            voting_power=voting_power,
            use_stored_data=use_stored_data,
        )

    @property
    def chain_params(self):
        if self.offline or self.rpc is None:
            return known_chains["HIVE"]
        else:
            return self.get_network()

    @property
    def hardfork(self):
        if self.offline or self.rpc is None:
            versions = known_chains["HIVE"]["min_version"]
        else:
            hf_prop = self.get_hardfork_properties()
            if "current_hardfork_version" in hf_prop:
                versions = hf_prop["current_hardfork_version"]
            else:
                versions = self.get_blockchain_version()
        return int(versions.split(".")[1])

    @property
    def is_hive(self):
        config = self.get_config()
        if config is None:
            return True
        return "HIVE_CHAIN_ID" in self.get_config()

    @property
    def hbd_symbol(self):
        """get the current chains symbol for HBD (e.g. "TBD" on testnet)"""
        return self.backed_token_symbol

    @property
    def hive_symbol(self):
        """get the current chains symbol for HIVE (e.g. "TESTS" on testnet)"""
        return self.token_symbol

    @property
    def vests_symbol(self):
        """get the current chains symbol for VESTS"""
        return self.vest_token_symbol



# mpl.use('Agg')
# mpl.use('TkAgg')
import matplotlib.pyplot as plt

from nectar import Steem
from nectar.amount import Amount
from nectar.utils import reputation_to_score

if __name__ == "__main__":
    stm = Steem()
    price = Amount(stm.get_current_median_history()["base"])
    reps = [0]
    for i in range(26, 91):
        reps.append(int(10 ** ((i - 25) / 9 + 9)))
    # reps = np.logspace(9, 16, 60)
    used_power = stm._calc_resulting_vote()
    last_sp = 0
    sp_list = []
    rep_score_list = []
    for goal_rep in reps:
        score = reputation_to_score(goal_rep)
        rep_score_list.append(score)
        needed_rshares = int(goal_rep) << 6
        needed_vests = needed_rshares / used_power / 100
        needed_sp = stm.vests_to_sp(needed_vests)
        sp_list.append(needed_sp / 1000)
        # print("| %.1f | %.2f | %.2f  | " % (score, needed_sp / 1000, needed_sp / 1000 - last_sp / 1000))
        last_sp = needed_sp

    plt.figure(figsize=(12, 6))
    opts = {"linestyle": "-", "marker": "."}
    plt.semilogx(sp_list, rep_score_list, label="Reputation", **opts)
    plt.grid()
    plt.legend()
    plt.title("Required number of 1k SP upvotes to reach certain reputation")
    plt.xlabel("1k SP votes")
    plt.ylabel("Reputation")
    plt.show()
    # plt.savefig("rep_based_on_votes.png")
